#  Copyright (c) 2020, Manfred Moitzi
#  License: MIT License
# Test only basic features of Cython implementation,
# Full testing and compatibility check with Python implementation
# is located in test suite 602.

import pytest

cyvec = pytest.importorskip("ezdxf.acc.vector")
Vec3 = cyvec.Vec3


def test_default_constructor():
    v = Vec3()
    assert v.x == 0
    assert v.y == 0
    assert v.z == 0


def test_is_immutable():
    v = Vec3()
    with pytest.raises(AttributeError):
        v.x = 1
    with pytest.raises(AttributeError):
        v.y = 1
    with pytest.raises(AttributeError):
        v.z = 1


def test_init_x_y():
    v = Vec3(1, 2, 3)
    assert v.x == 1
    assert v.y == 2
    assert v.z == 3


def test_init_2_tuple():
    v = Vec3((1, 2))
    assert v.x == 1
    assert v.y == 2
    assert v.z == 0


def test_init_vec2():
    v = Vec3(cyvec.Vec2(1, 2))
    assert v.x == 1
    assert v.y == 2
    assert v.z == 0


def test_init_3_tuple():
    v = Vec3((1, 2, 3))
    assert v.x == 1
    assert v.y == 2
    assert v.z == 3


def test_init_vec3():
    v = Vec3(Vec3(1, 2, 3))
    assert v.x == 1
    assert v.y == 2
    assert v.z == 3


def test_init_type_error():
    with pytest.raises(TypeError):
        Vec3(1)
    with pytest.raises(TypeError):
        Vec3(1, 2, 3, 4)
    with pytest.raises(TypeError):
        Vec3("xyz")
    with pytest.raises(TypeError):
        Vec3("xyz", "abc")


def test_is_null():
    assert Vec3().is_null is True


def test_bool():
    assert bool(Vec3(1, 0)) is True


def test_compare():
    assert Vec3(1, 2, 3) == Vec3(1, 2, 3)
    assert Vec3(1, 2, 3) == (1, 2, 3)
    assert (1, 2) == Vec3(1, 2)
    assert Vec3(1, 2, 3) != (
        1,
        2.0000000000001,
        3,
    ), "__eq__() should use the full floating point precision"
    assert Vec3(1, 2) != (2, 1)


def test_lower_than():
    assert Vec3(1, 2) < Vec3(2, 2)
    assert Vec3(2, 2) > Vec3(1, 2)


def test_does_not_support_slicing():
    with pytest.raises(TypeError):
        _ = Vec3(2, 1)[:]


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2019-2020 Manfred Moitzi
# License: MIT License
# created 2019-02-15
from typing import cast
import pytest

import ezdxf
from ezdxf.entities.image import Wipeout
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text
from ezdxf.layouts import VirtualLayout
from ezdxf.math import Vec2


@pytest.fixture
def doc():
    return ezdxf.new()


def test_wipeout_default_new():
    entity = Wipeout.new(
        handle="ABBA", owner="0", dxfattribs={"image_size": (640, 200)}
    )
    assert entity.dxf.layer == "0"
    assert entity.dxf.insert is None  # set by add_image()
    assert entity.dxf.u_pixel is None  # set by add_image()
    assert entity.dxf.v_pixel is None  # set by add_image()
    assert entity.dxf.class_version == 0
    assert entity.dxf.image_size == (640, 200)
    assert entity.dxf.image_def_handle == "0"
    assert entity.dxf.flags == 3
    assert entity.dxf.clipping == 0
    assert entity.dxf.brightness == 50
    assert entity.dxf.contrast == 50
    assert entity.dxf.fade == 0
    assert entity.dxf.image_def_reactor_handle == "0"
    assert entity.dxf.clipping_boundary_type == 1
    assert entity.dxf.clip_mode == 0
    assert entity.boundary_path[0] == (-0.5, -0.5)
    assert entity.boundary_path[1] == (639.5, 199.5)


def test_wipeout_write_dxf():
    entity = Wipeout.from_text(WIPEOUT)
    result = TagCollector.dxftags(entity)
    expected = basic_tags_from_text(WIPEOUT)
    assert result == expected


def test_wipeout_creator_interface():
    layout = VirtualLayout()
    wipeout = layout.add_wipeout(
        [(150, 100), (250, 100), (250, 200), (150, 200)]
    )
    assert wipeout.dxftype() == "WIPEOUT"
    assert wipeout.dxf.insert == (150, 100, 0)
    assert wipeout.dxf.u_pixel == (100, 0, 0)
    assert wipeout.dxf.v_pixel == (0, 100, 0)
    assert wipeout.dxf.image_size == (1, 1)
    assert wipeout.dxf.flags == 7
    assert wipeout.dxf.clipping == 1
    assert wipeout.dxf.clip_mode == 0

    path = wipeout.boundary_path
    assert path[0].isclose(Vec2(-0.5, 0.5))
    assert path[1].isclose(Vec2(0.5, 0.5))
    assert path[2].isclose(Vec2(0.5, -0.5))
    assert path[3].isclose(Vec2(-0.5, -0.5))
    assert path[4] == path[0]

    path = wipeout.boundary_path_wcs()
    assert path[0] == (150, 100)
    assert path[1] == (250, 100)
    assert path[2] == (250, 200)
    assert path[3] == (150, 200)
    assert path[0] == path[-1]


def test_copy_wipeout():
    layout = VirtualLayout()
    wipeout = layout.add_wipeout([(150, 100), (250, 200)])
    copy = wipeout.copy()
    layout.add_entity(copy)
    assert len(layout) == 2


def test_wipeout_vars(doc):
    doc.set_wipeout_variables(frame=1)
    wipeout_variables = doc.rootdict["ACAD_WIPEOUT_VARS"]
    assert wipeout_variables.dxf.frame == 1


def test_boundary_path_wcs():
    from ezdxf.layouts import VirtualLayout

    layout = VirtualLayout()
    e = cast(
        Wipeout,
        layout.new_entity(
            "WIPEOUT",
            dxfattribs={
                "layer": "0",
                "class_version": 0,
                "insert": (150.0, 100.0, 0.0),
                "u_pixel": (100.0, 0.0, 0.0),
                "v_pixel": (0.0, 100.0, 0.0),
                "image_size": (1.0, 1.0, 0.0),
                "image_def_handle": "0",
                "flags": 7,
                "clipping": 1,
                "brightness": 50,
                "contrast": 50,
                "fade": 0,
                "image_def_reactor_handle": "0",
                "clipping_boundary_type": 2,
                "count_boundary_points": 5,
                "clip_mode": 0,
            },
        ),
    )
    e.set_boundary_path(
        [
            (-0.5, 0.5),
            (0.5, 0.5),
            (0.5, -0.5),
            (-0.5, -0.5),
            (-0.5, 0.5),
        ]
    )
    path = e.boundary_path_wcs()
    assert path[0] == (150, 100)
    assert path[1] == (250, 100)
    assert path[2] == (250, 200)
    assert path[3] == (150, 200)
    assert path[0] == path[-1]


WIPEOUT = """0
WIPEOUT
5
0
330
0
100
AcDbEntity
8
0
100
AcDbWipeout
90
0
10
0.0
20
0.0
30
0.0
11
0.0
21
0.0
31
0.0
12
0.0
22
0.0
32
0.0
13
640
23
320
340
0
70
3
280
1
281
50
282
50
283
0
360
0
71
1
91
2
14
-0.5
24
-0.5
14
639.5
24
319.5
290
0
"""



#  Copyright (c) 2022-2023, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import (
    Iterator,
    Sequence,
    TYPE_CHECKING,
    Callable,
    Any,
    Union,
    Optional,
    Tuple,
)
from typing_extensions import TypeAlias
from collections import defaultdict
import enum
import math
import dataclasses
import random
from ezdxf.math import (
    Vec2,
    Vec3,
    Bezier3P,
    Bezier4P,
    intersection_ray_cubic_bezier_2d,
    quadratic_to_cubic_bezier,
)
from ezdxf import const
from ezdxf.path import Path, LineTo, MoveTo, Curve3To, Curve4To

if TYPE_CHECKING:
    from ezdxf.entities.polygon import DXFPolygon

MIN_HATCH_LINE_DISTANCE = 1e-4  # ??? what's a good choice
NONE_VEC2 = Vec2(math.nan, math.nan)
KEY_NDIGITS = 4
SORT_NDIGITS = 10


class IntersectionType(enum.IntEnum):
    NONE = 0
    REGULAR = 1
    START = 2
    END = 3
    COLLINEAR = 4


class HatchingError(Exception):
    """Base exception class of the :mod:`hatching` module."""

    pass


class HatchLineDirectionError(HatchingError):
    """Hatching direction is undefined or a (0, 0) vector."""

    pass


class DenseHatchingLinesError(HatchingError):
    """Very small hatching distance which creates too many hatching lines."""

    pass


@dataclasses.dataclass(frozen=True)
class Line:
    start: Vec2
    end: Vec2
    distance: float  # normal distance to the hatch baseline


@dataclasses.dataclass(frozen=True)
class Intersection:
    """Represents an intersection."""

    type: IntersectionType = IntersectionType.NONE
    p0: Vec2 = NONE_VEC2
    p1: Vec2 = NONE_VEC2


def side_of_line(distance: float, abs_tol=1e-12) -> int:
    if abs(distance) < abs_tol:
        return 0
    if distance > 0.0:
        return +1
    return -1


@dataclasses.dataclass(frozen=True)
class HatchLine:
    """Represents a single hatch line.

    Args:
        origin: the origin of the hatch line as :class:`~ezdxf.math.Vec2` instance
        direction: the hatch line direction as :class:`~ezdxf.math.Vec2` instance, must not (0, 0)
        distance: the normal distance to the base hatch line as float

    """

    origin: Vec2
    direction: Vec2
    distance: float

    def intersect_line(
        self,
        a: Vec2,
        b: Vec2,
        dist_a: float,
        dist_b: float,
    ) -> Intersection:
        """Returns the :class:`Intersection` of this hatch line and the line
        defined by the points `a` and `b`.
        The arguments `dist_a` and `dist_b` are the signed normal distances of
        the points `a` and `b` from the hatch baseline.
        The normal distances from the baseline are easy to calculate by the
        :meth:`HatchBaseLine.signed_distance` method and allow a fast
        intersection calculation by a simple point interpolation.

        Args:
            a: start point of the line as :class:`~ezdxf.math.Vec2` instance
            b: end point of the line as :class:`~ezdxf.math.Vec2` instance
            dist_a: normal distance of point `a` to the hatch baseline as float
            dist_b: normal distance of point `b` to the hatch baseline as float

        """
        # all distances are normal distances to the hatch baseline
        line_distance = self.distance
        side_a = side_of_line(dist_a - line_distance)
        side_b = side_of_line(dist_b - line_distance)
        if side_a == 0:
            if side_b == 0:
                return Intersection(IntersectionType.COLLINEAR, a, b)
            else:
                return Intersection(IntersectionType.START, a)
        elif side_b == 0:
            return Intersection(IntersectionType.END, b)
        elif side_a != side_b:
            factor = abs((dist_a - line_distance) / (dist_a - dist_b))
            return Intersection(IntersectionType.REGULAR, a.lerp(b, factor))
        return Intersection()  # no intersection

    def intersect_cubic_bezier_curve(self, curve: Bezier4P) -> Sequence[Intersection]:
        """Returns 0 to 3 :class:`Intersection` points of this hatch line with
        a cubic Bèzier curve.

        Args:
            curve: the cubic Bèzier curve as :class:`ezdxf.math.Bezier4P` instance

        """
        return [
            Intersection(IntersectionType.REGULAR, p, NONE_VEC2)
            for p in intersection_ray_cubic_bezier_2d(
                self.origin, self.origin + self.direction, curve
            )
        ]


class PatternRenderer:
    """
    The hatch pattern of a DXF entity has one or more :class:`HatchBaseLine`
    instances with an origin, direction, offset and line pattern.
    The :class:`PatternRenderer` for a certain distance from the
    baseline has to be acquired from the :class:`HatchBaseLine` by the
    :meth:`~HatchBaseLine.pattern_renderer` method.

    The origin of the hatch line is the starting point of the line
    pattern. The offset defines the origin of the adjacent
    hatch line and doesn't have to be orthogonal to the hatch line direction.

    **Line Pattern**

    The line pattern is a sequence of floats, where a value > 0.0 is a dash, a
    value < 0.0 is a gap and value of 0.0 is a point.

    Args:
        hatch_line: :class:`HatchLine`
        pattern: the line pattern as sequence of float values

    """

    def __init__(self, hatch_line: HatchLine, pattern: Sequence[float]):
        self.origin = hatch_line.origin
        self.direction = hatch_line.direction
        self.pattern = pattern
        self.pattern_length = math.fsum([abs(e) for e in pattern])

    def sequence_origin(self, index: float) -> Vec2:
        return self.origin + self.direction * (self.pattern_length * index)

    def render(self, start: Vec2, end: Vec2) -> Iterator[tuple[Vec2, Vec2]]:
        """Yields the pattern lines as pairs of :class:`~ezdxf.math.Vec2`
        instances from the start- to the end point on the hatch line.
        For points the start- and end point are the same :class:`~ezdxf.math.Vec2`
        instance and can be tested by the ``is`` operator.

        The start- and end points should be located collinear at the hatch line
        of this instance, otherwise the points a projected onto this hatch line.

        """
        if start.isclose(end):
            return
        length = self.pattern_length
        if length < 1e-9:
            yield start, end
            return

        direction = self.direction
        if direction.dot(end - start) < 0.0:
            # Line direction is reversed to the pattern line direction!
            start, end = end, start
        origin = self.origin
        s_dist = direction.dot(start - origin)
        e_dist = direction.dot(end - origin)
        s_index, s_offset = divmod(s_dist, length)
        e_index, e_offset = divmod(e_dist, length)

        if s_index == e_index:
            yield from self.render_offset_to_offset(s_index, s_offset, e_offset)
            return
        # line crosses pattern border
        if s_offset > 0.0:
            yield from self.render_offset_to_offset(
                s_index,
                s_offset,
                length,
            )
            s_index += 1

        while s_index < e_index:
            yield from self.render_full_pattern(s_index)
            s_index += 1

        if e_offset > 0.0:
            yield from self.render_offset_to_offset(
                s_index,
                0.0,
                e_offset,
            )

    def render_full_pattern(self, index: float) -> Iterator[tuple[Vec2, Vec2]]:
        # fast pattern rendering
        direction = self.direction
        start_point = self.sequence_origin(index)
        for dash in self.pattern:
            if dash == 0.0:
                yield start_point, start_point
            else:
                end_point = start_point + direction * abs(dash)
                if dash > 0.0:
                    yield start_point, end_point
                start_point = end_point

    def render_offset_to_offset(
        self, index: float, s_offset: float, e_offset: float
    ) -> Iterator[tuple[Vec2, Vec2]]:
        direction = self.direction
        origin = self.sequence_origin(index)
        start_point = origin + direction * s_offset
        distance = 0.0
        for dash in self.pattern:
            distance += abs(dash)
            if distance < s_offset:
                continue
            if dash == 0.0:
                yield start_point, start_point
            else:
                end_point = origin + direction * min(distance, e_offset)
                if dash > 0.0:
                    yield start_point, end_point
                if distance >= e_offset:
                    return
                start_point = end_point


class HatchBaseLine:
    """A hatch baseline defines the source line for hatching a geometry.
    A complete hatch pattern of a DXF entity can consist of one or more hatch
    baselines.

    Args:
        origin: the origin of the hatch line as :class:`~ezdxf.math.Vec2` instance
        direction: the hatch line direction as :class:`~ezdxf.math.Vec2` instance, must not (0, 0)
        offset: the offset of the hatch line origin to the next or to the previous hatch line
        line_pattern: line pattern as sequence of floats, see also :class:`PatternRenderer`
        min_hatch_line_distance: minimum hatch line distance to render, raises an
            :class:`DenseHatchingLinesError` exception if the distance between hatch
            lines is smaller than this value

    Raises:
        HatchLineDirectionError: hatch baseline has no direction, (0, 0) vector
        DenseHatchingLinesError: hatching lines are too narrow

    """

    def __init__(
        self,
        origin: Vec2,
        direction: Vec2,
        offset: Vec2,
        line_pattern: Optional[list[float]] = None,
        min_hatch_line_distance=MIN_HATCH_LINE_DISTANCE,
    ):
        self.origin = origin
        try:
            self.direction = direction.normalize()
        except ZeroDivisionError:
            raise HatchLineDirectionError("hatch baseline has no direction")
        self.offset = offset
        self.normal_distance: float = (-offset).det(self.direction - offset)
        if abs(self.normal_distance) < min_hatch_line_distance:
            raise DenseHatchingLinesError("hatching lines are too narrow")
        self._end = self.origin + self.direction
        self.line_pattern: list[float] = line_pattern if line_pattern else []

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(origin={self.origin!r}, "
            f"direction={self.direction!r}, offset={self.offset!r})"
        )

    def hatch_line(self, distance: float) -> HatchLine:
        """Returns the :class:`HatchLine` at the given signed `distance`."""
        factor = distance / self.normal_distance
        return HatchLine(self.origin + self.offset * factor, self.direction, distance)

    def signed_distance(self, point: Vec2) -> float:
        """Returns the signed normal distance of the given `point` from this
        hatch baseline.
        """
        # denominator (_end - origin).magnitude is 1.0 !!!
        return (self.origin - point).det(self._end - point)

    def pattern_renderer(self, distance: float) -> PatternRenderer:
        """Returns the :class:`PatternRenderer` for the given signed `distance`."""
        return PatternRenderer(self.hatch_line(distance), self.line_pattern)


def hatch_line_distances(
    point_distances: Sequence[float], normal_distance: float
) -> list[float]:
    """Returns all hatch line distances in the range of the given point
    distances.
    """
    assert normal_distance != 0.0
    normal_factors = [d / normal_distance for d in point_distances]
    max_line_number = int(math.ceil(max(normal_factors)))
    min_line_number = int(math.ceil(min(normal_factors)))
    return [normal_distance * num for num in range(min_line_number, max_line_number)]


def intersect_polygon(
    baseline: HatchBaseLine, polygon: Sequence[Vec2]
) -> Iterator[tuple[Intersection, float]]:
    """Yields all intersection points of the hatch defined by the `baseline` and
    the given `polygon`.

    Returns the intersection point and the normal-distance from the baseline,
    intersection points with the same normal-distance lay on the same hatch
    line.

    """
    count = len(polygon)
    if count < 3:
        return
    if polygon[0].isclose(polygon[-1]):
        count -= 1
        if count < 3:
            return

    prev_point = polygon[count - 1]  # last point
    dist_prev = baseline.signed_distance(prev_point)
    for index in range(count):
        point = polygon[index]
        dist_point = baseline.signed_distance(point)
        for hatch_line_distance in hatch_line_distances(
            (dist_prev, dist_point), baseline.normal_distance
        ):
            hatch_line = baseline.hatch_line(hatch_line_distance)
            ip = hatch_line.intersect_line(
                prev_point,
                point,
                dist_prev,
                dist_point,
            )
            if (
                ip.type != IntersectionType.NONE
                and ip.type != IntersectionType.COLLINEAR
            ):
                yield ip, hatch_line_distance

        prev_point = point
        dist_prev = dist_point


def hatch_polygons(
    baseline: HatchBaseLine,
    polygons: Sequence[Sequence[Vec2]],
    terminate: Optional[Callable[[], bool]] = None,
) -> Iterator[Line]:
    """Yields all pattern lines for all hatch lines generated by the given
    :class:`HatchBaseLine`, intersecting the given 2D polygons as :class:`Line`
    instances.
    The `polygons` should represent a single entity with or without holes, the
    order of the polygons and their winding orientation (cw or ccw) is not
    important. Entities which do not intersect or overlap should be handled
    separately!

    Each polygon is a sequence of :class:`~ezdxf.math.Vec2` instances, they are
    treated as closed polygons even if the last vertex is not equal to the
    first vertex.

    The hole detection is done by a simple inside/outside counting algorithm and
    far from perfect, but is able to handle ordinary polygons well.

    The terminate function WILL BE CALLED PERIODICALLY AND should return
    ``True`` to terminate execution. This can be used to implement a timeout,
    which can be required if using a very small hatching distance, especially
    if you get the data from untrusted sources.

    Args:
        baseline: :class:`HatchBaseLine`
        polygons: multiple sequences of :class:`~ezdxf.path.Vec2` instances of
            a single entity, the order of exterior- and hole paths and the
            winding orientation (cw or ccw) of paths is not important
        terminate: callback function which is called periodically and should
            return ``True`` to terminate the hatching function

    """
    yield from _hatch_geometry(baseline, polygons, intersect_polygon, terminate)


def intersect_path(
    baseline: HatchBaseLine, path: Path
) -> Iterator[tuple[Intersection, float]]:
    """Yields all intersection points of the hatch defined by the `baseline` and
    the given single `path`.

    Returns the intersection point and the normal-distance from the baseline,
    intersection points with the same normal-distance lay on the same hatch
    line.

    """
    for path_element in _path_elements(path):
        if isinstance(path_element, Bezier4P):
            distances = [
                baseline.signed_distance(p) for p in path_element.control_points
            ]
            for hatch_line_distance in hatch_line_distances(
                distances, baseline.normal_distance
            ):
                hatch_line = baseline.hatch_line(hatch_line_distance)
                for ip in hatch_line.intersect_cubic_bezier_curve(path_element):
                    yield ip, hatch_line_distance
        else:  # line
            a, b = Vec2.generate(path_element)
            dist_a = baseline.signed_distance(a)
            dist_b = baseline.signed_distance(b)
            for hatch_line_distance in hatch_line_distances(
                (dist_a, dist_b), baseline.normal_distance
            ):
                hatch_line = baseline.hatch_line(hatch_line_distance)
                ip = hatch_line.intersect_line(a, b, dist_a, dist_b)
                if (
                    ip.type != IntersectionType.NONE
                    and ip.type != IntersectionType.COLLINEAR
                ):
                    yield ip, hatch_line_distance


def _path_elements(path: Path) -> Union[Bezier4P, tuple[Vec2, Vec2]]:
    if len(path) == 0:
        return
    start = path.start
    path_start = start
    for command in path.commands():
        end = command.end
        if isinstance(command, MoveTo):
            if not path_start.isclose(start):
                yield start, path_start  # close sub-path
            path_start = end
        elif isinstance(command, LineTo) and not start.isclose(end):
            yield start, end
        elif isinstance(command, Curve4To):
            yield Bezier4P((start, command.ctrl1, command.ctrl2, end))
        elif isinstance(command, Curve3To):
            curve3 = Bezier3P((start, command.ctrl, end))
            yield quadratic_to_cubic_bezier(curve3)
        start = end

    if not path_start.isclose(start):  # close path
        yield start, path_start


def hatch_paths(
    baseline: HatchBaseLine,
    paths: Sequence[Path],
    terminate: Optional[Callable[[], bool]] = None,
) -> Iterator[Line]:
    """Yields all pattern lines for all hatch lines generated by the given
    :class:`HatchBaseLine`, intersecting the given 2D :class:`~ezdxf.path.Path`
    instances as :class:`Line` instances. The paths are handled as projected
    into the xy-plane the z-axis of path vertices will be ignored if present.

    Same as the :func:`hatch_polygons` function, but for :class:`~ezdxf.path.Path`
    instances instead of polygons build of vertices. This function **does not
    flatten** the paths into vertices, instead the real intersections of the
    Bézier curves and the hatch lines are calculated.

    For more information see the docs of the :func:`hatch_polygons` function.

    Args:
        baseline: :class:`HatchBaseLine`
        paths: sequence of :class:`~ezdxf.path.Path` instances of a single
            entity, the order of exterior- and hole paths and the winding
            orientation (cw or ccw) of the paths is not important
        terminate: callback function which is called periodically and should
            return ``True`` to terminate the hatching function

    """
    yield from _hatch_geometry(baseline, paths, intersect_path, terminate)


IFuncType: TypeAlias = Callable[
    [HatchBaseLine, Any], Iterator[Tuple[Intersection, float]]
]


def _hatch_geometry(
    baseline: HatchBaseLine,
    geometries: Sequence[Any],
    intersection_func: IFuncType,
    terminate: Optional[Callable[[], bool]] = None,
) -> Iterator[Line]:
    """Returns all pattern lines intersecting the given geometries.

    The intersection_func() should yield all intersection points between a
    HatchBaseLine() and as given geometry.

    The terminate function should return ``True`` to terminate execution
    otherwise ``False``. Can be used to implement a timeout.

    """
    points: dict[float, list[Intersection]] = defaultdict(list)
    for geometry in geometries:
        if terminate and terminate():
            return
        for ip, distance in intersection_func(baseline, geometry):
            assert ip.type != IntersectionType.NONE
            points[round(distance, KEY_NDIGITS)].append(ip)

    for distance, vertices in points.items():
        if terminate and terminate():
            return
        start = NONE_VEC2
        end = NONE_VEC2
        for line in _line_segments(vertices, distance):
            if start is NONE_VEC2:
                start = line.start
                end = line.end
                continue
            if line.start.isclose(end):
                end = line.end
            else:
                yield Line(start, end, distance)
                start = line.start
                end = line.end

        if start is not NONE_VEC2:
            yield Line(start, end, distance)


def _line_segments(vertices: list[Intersection], distance: float) -> Iterator[Line]:
    if len(vertices) < 2:
        return
    vertices.sort(key=lambda p: p.p0.round(SORT_NDIGITS))
    inside = False
    prev_point = NONE_VEC2
    for ip in vertices:
        if ip.type == IntersectionType.NONE or ip.type == IntersectionType.COLLINEAR:
            continue
        # REGULAR, START, END
        point = ip.p0
        if prev_point is NONE_VEC2:
            inside = True
            prev_point = point
            continue
        if inside:
            yield Line(prev_point, point, distance)

        inside = not inside
        prev_point = point


def hatch_entity(
    polygon: DXFPolygon,
    filter_text_boxes=True,
    jiggle_origin: bool = True,
) -> Iterator[tuple[Vec3, Vec3]]:
    """Yields the hatch pattern of the given HATCH or MPOLYGON entity as 3D lines.
    Each line is a pair of :class:`~ezdxf.math.Vec3` instances as start- and end
    vertex, points are represented as lines of zero length, which means the
    start vertex is equal to the end vertex.

    The function yields nothing if `polygon` has a solid- or gradient filling
    or does not have a usable pattern assigned.

    Args:
        polygon: :class:`~ezdxf.entities.Hatch` or :class:`~ezdxf.entities.MPolygon`
            entity
        filter_text_boxes: ignore text boxes if ``True``
        jiggle_origin: move pattern line origins a small amount to avoid intersections
            in corner points which causes errors in patterns

    """
    if polygon.pattern is None or polygon.dxf.solid_fill:
        return
    if len(polygon.pattern.lines) == 0:
        return
    ocs = polygon.ocs()
    elevation = polygon.dxf.elevation.z
    paths = hatch_boundary_paths(polygon, filter_text_boxes)
    # todo: MPOLYGON offset
    # All paths in OCS!
    for baseline in pattern_baselines(polygon, jiggle_origin=jiggle_origin):
        for line in hatch_paths(baseline, paths):
            line_pattern = baseline.pattern_renderer(line.distance)
            for s, e in line_pattern.render(line.start, line.end):
                if ocs.transform:
                    yield ocs.to_wcs((s.x, s.y, elevation)), ocs.to_wcs(
                        (e.x, e.y, elevation)
                    )
                yield Vec3(s), Vec3(e)


def hatch_boundary_paths(polygon: DXFPolygon, filter_text_boxes=True) -> list[Path]:
    """Returns the hatch boundary paths as :class:`ezdxf.path.Path` instances
    of HATCH and MPOLYGON entities. Ignores text boxes if argument
    `filter_text_boxes` is ``True``.
    """
    from ezdxf.path import from_hatch_boundary_path

    loops = []
    for boundary in polygon.paths.rendering_paths(polygon.dxf.hatch_style):
        if filter_text_boxes and boundary.path_type_flags & const.BOUNDARY_PATH_TEXTBOX:
            continue
        path = from_hatch_boundary_path(boundary)
        for sub_path in path.sub_paths():
            if len(sub_path):
                sub_path.close()
                loops.append(sub_path)
    return loops


def _jiggle_factor():
    # range 0.0003 .. 0.0010
    return random.random() * 0.0007 + 0.0003


def pattern_baselines(
    polygon: DXFPolygon,
    min_hatch_line_distance: float = MIN_HATCH_LINE_DISTANCE,
    *,
    jiggle_origin: bool = False,
) -> Iterator[HatchBaseLine]:
    """Yields the hatch pattern baselines of HATCH and MPOLYGON entities as
    :class:`HatchBaseLine` instances.  Set `jiggle_origin` to ``True`` to move pattern
    line origins a small amount to avoid intersections in corner points which causes
    errors in patterns.

    """
    pattern = polygon.pattern
    if not pattern:
        return
    # The hatch pattern parameters are already scaled and rotated for direct
    # usage!
    # The stored scale and angle is just for reconstructing the base pattern
    # when applying a new scaling or rotation.

    jiggle_offset = Vec2()
    if jiggle_origin:
        # move origin of base pattern lines a small amount to avoid intersections with
        # boundary corner points
        offsets: list[float] = [line.offset.magnitude for line in pattern.lines]
        if len(offsets):
            # calculate the same random jiggle offset for all pattern base lines
            mean = sum(offsets) / len(offsets)
            x = _jiggle_factor() * mean
            y = _jiggle_factor() * mean
            jiggle_offset = Vec2(x, y)

    for line in pattern.lines:
        direction = Vec2.from_deg_angle(line.angle)
        yield HatchBaseLine(
            origin=line.base_point + jiggle_offset,
            direction=direction,
            offset=line.offset,
            line_pattern=line.dash_length_items,
            min_hatch_line_distance=min_hatch_line_distance,
        )



#  Copyright (c) 2022, Manfred Moitzi
#  License: MIT License
from typing import NamedTuple, Union
import os
from ezdxf.lldxf import const


class DWGInfo(NamedTuple):
    version: str = "unknown"
    release: str = "unknown"


def dwg_info(data: bytes) -> DWGInfo:
    """Returns the version and release name of a DWG file."""
    if len(data) < 6:
        return DWGInfo("invalid", "invalid")
    version = data[:6].decode(errors="ignore")
    if version[:4] != "AC10":
        return DWGInfo("invalid", "invalid")
    release = const.acad_release.get(version, "unknown")
    return DWGInfo(version, release)


def dwg_file_info(file: Union[str, os.PathLike]) -> DWGInfo:
    """Returns the version and release name of a DWG file."""
    with open(file, "rb") as fp:
        return dwg_info(fp.read(6))



# Copyright (c) 2019 Manfred Moitzi
# License: MIT License
# created: 2019-02-18
import pytest

from ezdxf.document import Drawing
from ezdxf.sections.objects import _OBJECT_TABLE_NAMES


@pytest.fixture
def doc():
    return Drawing.new("r2018")


def test_create_new_empty_drawing(doc):
    assert doc.dxfversion == "AC1032"
    rootdict = doc.rootdict
    assert rootdict.DXFTYPE == "DICTIONARY"
    for name in _OBJECT_TABLE_NAMES:
        assert name in rootdict


def test_section(doc):
    assert doc.header["$ACADVER"] == "AC1032"

    e = doc.objects.get_entity_space()
    assert e[0] is doc.rootdict

    assert len(doc.blocks) == 2
    assert "*Model_Space" in doc.blocks
    assert "*Paper_Space" in doc.blocks


def test_viewports(doc):
    assert len(doc.viewports) == 1
    # viewport table can have multiple entries with same name, therefore returns a list
    assert len(doc.viewports.get("*Active")) == 1

    assert len(doc.linetypes) == 3


def test_layers(doc):
    assert len(doc.layers) == 2
    assert "0" in doc.layers
    assert "Defpoints" in doc.layers
    assert (
        doc.layers.get("0").dxf.material_handle
        == doc.materials["Global"].dxf.handle
    )
    assert (
        doc.layers.get("0").dxf.plotstyle_handle
        == doc.plotstyles["Normal"].dxf.handle
    )
    assert len(doc.views) == 0


def test_linetypes(doc):
    assert len(doc.linetypes) == 3
    assert "ByLayer" in doc.linetypes
    assert "ByBlock" in doc.linetypes
    assert "Continuous" in doc.linetypes


def test_text_styles(doc):
    assert len(doc.styles) == 1
    assert "Standard" in doc.styles


def test_dim_styles(doc):
    assert len(doc.dimstyles) == 1
    assert "Standard" in doc.dimstyles


def test_views(doc):
    assert len(doc.views) == 0


def test_appids(doc):
    assert len(doc.appids) == 1


def test_ucs(doc):
    assert len(doc.ucs) == 0


def test_get_modelspace(doc):
    msp = doc.modelspace()
    assert len(msp) == 0
    msp.add_line((0, 0), (1, 1))
    assert len(msp) == 1


def test_get_paperspace(doc):
    psp = doc.layout()
    assert len(psp) == 0
    psp.add_line((0, 0), (1, 1))
    assert len(psp) == 1



# Copyright (c) 2019 Manfred Moitzi
# License: MIT License
import pytest
import ezdxf
from ezdxf.entities.tolerance import Tolerance
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text

TOLERANCE = """0
TOLERANCE
5
0
330
0
100
AcDbEntity
8
0
100
AcDbFcf
3
Standard
10
0.0
20
0.0
30
0.0
1

11
1.0
21
0.0
31
0.0
"""


@pytest.fixture
def entity():
    return Tolerance.from_text(TOLERANCE)


def test_registered():
    from ezdxf.entities.factory import ENTITY_CLASSES

    assert "TOLERANCE" in ENTITY_CLASSES


def test_default_init():
    entity = Tolerance()
    assert entity.dxftype() == "TOLERANCE"
    assert entity.dxf.handle is None
    assert entity.dxf.owner is None


def test_default_new():
    entity = Tolerance.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "color": 7,
            "dimstyle": "EZDXF",
            "insert": (1, 2, 3),
            "extrusion": (4, 5, 6),
            "x_axis_vector": (7, 8, 9),
            "content": "abcdef",
        },
    )
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 7
    assert entity.dxf.dimstyle == "EZDXF"
    assert entity.dxf.insert == (1, 2, 3)
    assert entity.dxf.extrusion == (4, 5, 6)
    assert entity.dxf.x_axis_vector == (7, 8, 9)
    assert entity.dxf.content == "abcdef"


def test_load_from_text(entity):
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 256, "default color is 256 (by layer)"
    assert entity.dxf.dimstyle == "Standard"
    assert entity.dxf.insert == (0, 0, 0)
    assert entity.dxf.extrusion == (0, 0, 1)  # default value
    assert entity.dxf.x_axis_vector == (1, 0, 0)
    assert entity.dxf.content == ""


def test_write_dxf():
    entity = Tolerance.from_text(TOLERANCE)
    result = TagCollector.dxftags(entity)
    expected = basic_tags_from_text(TOLERANCE)
    assert result == expected


def test_add_tolerance():
    doc = ezdxf.new()
    msp = doc.modelspace()
    light = msp.new_entity("TOLERANCE", {})
    assert light.dxftype() == "TOLERANCE"
    assert light.dxf.dimstyle == "Standard"
    assert light.dxf.insert == (0, 0, 0)
    assert light.dxf.content == ""
    assert light.dxf.extrusion == (0, 0, 1)
    assert light.dxf.x_axis_vector == (1, 0, 0)



#  Copyright (c) 2022, Manfred Moitzi
#  License: MIT License
from typing import Sequence, Iterable, List
import os
import json
from pathlib import Path
from functools import partial
import pytest
from ezdxf.math import Vec2, UVec
from ezdxf.math._mapbox_earcut import earcut as _py_earcut

CYTHON = "Cython"
try:
    from ezdxf.acc.mapbox_earcut import earcut as _cy_earcut
except ImportError:
    CYTHON = "CPython"
    _cy_earcut = _py_earcut


def earcut_driver(
    exterior: Iterable[UVec],
    holes: Iterable[Iterable[UVec]] = None,
    func=_py_earcut,
) -> List[Sequence[Vec2]]:
    points: List[Vec2] = Vec2.list(exterior)
    if len(points) == 0:
        return []
    holes_: List[List[Vec2]] = []
    if holes:
        holes_ = [Vec2.list(hole) for hole in holes]
    return func(points, holes_)


py_earcut = partial(earcut_driver, func=_py_earcut)
cy_earcut = partial(earcut_driver, func=_cy_earcut)


BASEDIR = os.path.dirname(__file__)
DATADIR = "mapbox-test-data"

EXPECTED = {
    "bad-diagonals": 7,
    "bad-hole": 42,
    "boxy": 57,
    "building": 13,
    "collinear-diagonal": 14,
    "degenerate": 0,
    "dude": 106,
    "eberly-3": 73,
    "eberly-6": 1429,
    "empty-square": 0,
    "filtered-bridge-jhl": 25,
    "hilbert": 1024,
    "hole-touching-outer": 77,
    "hourglass": 2,
    "infinite-loop-jhl": 0,
    "issue107": 0,
    # "issue111": 19,  # error? - I don't see it!
    "issue119": 18,
    "issue131": 12,
    # "issue142": 4, disabled because of issue #713 on some aarch64 systems
    "issue149": 2,
    "issue16": 12,
    "issue17": 11,
    "issue29": 40,
    "issue34": 139,
    "issue35": 844,
    "issue45": 10,
    "issue52": 109,
    "issue83": 0,
    "outside-ring": 64,
    "rain": 2681,
    "self-touching": 124,
    "shared-points": 4,
    "simplified-us-border": 120,
    "steiner": 9,
    "touching-holes": 57,
    "touching2": 8,
    "touching3": 15,
    # "touching4": 20,  # error? - I don't see it!
    "water": 2482,
    # "water-huge": 5177,  # long running test
    # "water-huge2": 4462,  # long running test
    "water2": 1212,
    "water3": 197,
    "water3b": 25,
    "water4": 705,
}

FILEPATHS = list(
    p for p in (Path(BASEDIR) / DATADIR).glob("*.json") if p.stem in EXPECTED
)


@pytest.fixture(params=FILEPATHS, ids=[p.stem for p in FILEPATHS])
def filename(request):
    return request.param


@pytest.mark.parametrize(
    "earcut", [py_earcut, cy_earcut], ids=("CPython", CYTHON)
)
def test_mapbox_earcut(filename: Path, earcut):
    name = filename.stem
    with filename.open("rt") as fp:
        data = json.load(fp)
        shape = data[0]
        holes = data[1:]
        triangles = earcut(shape, holes)
        assert len(triangles) == EXPECTED[name], f"{name}.json failed"


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License

import pytest
from pathlib import Path
from xml.etree import ElementTree as ET
from ezdxf.addons.hpgl2 import api as hpgl2

DATA = Path(__file__).parent / "data"
PLOTFILE = DATA / "PLOTFILE.plt"

if not PLOTFILE.exists():
    pytest.skip(reason=f"required file '{PLOTFILE}' not found", allow_module_level=True)


@pytest.fixture(scope="module")
def svg_str():
    data = PLOTFILE.read_bytes()
    return hpgl2.to_svg(data)


def test_svg_was_created(svg_str):
    assert len(svg_str) > 70000


def test_basic_svg_attributes(svg_str):
    root = ET.fromstring(svg_str)
    assert root.tag.endswith("svg")
    assert root.attrib["width"] == "593.8mm"
    assert root.attrib["height"] == "419.9mm"
    assert root.attrib["viewBox"] == "0 0 23752 16794"


if __name__ == '__main__':
    pytest.main([__file__])



#  Copyright (c) 2022-2024, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from dataclasses import dataclass, field
from datetime import datetime
import struct
from . import const

# ACIS versions exported by BricsCAD:
# R2000/AC1015: 400, "ACIS 4.00 NT", text length has no prefix "@"
# R2004/AC1018: 20800 @ "ACIS 208.00 NT", text length has "@" prefix ??? weird
# R2007/AC1021: 700 @ "ACIS 32.0 NT", text length has "@" prefix
# R2010/AC1024: 700 @ "ACIS 32.0 NT", text length has "@" prefix

# A test showed that R2000 files that contains ACIS v700/32.0 or v20800/208.0
# data can be opened by Autodesk TrueView, BricsCAD and Allplan, so exporting
# only v700/32.0 for all DXF versions should be OK!
# test script: exploration/acis/transplant_acis_data.py


def encode_str(s: str) -> bytes:
    b = s.encode("utf8", errors="ignore")
    return struct.pack("<BB", const.Tags.STR, len(b)) + b


def encode_double(value: float) -> bytes:
    return struct.pack("<Bd", const.Tags.DOUBLE, value)


@dataclass
class AcisHeader:
    """Represents an ACIS file header."""

    version: int = const.MIN_EXPORT_VERSION
    n_records: int = 0  # can be 0
    n_entities: int = 0
    flags: int = 0
    product_id: str = const.EZDXF_BUILDER_ID
    acis_version: str = const.ACIS_VERSION[const.MIN_EXPORT_VERSION]
    creation_date: datetime = field(default_factory=datetime.now)
    units_in_mm: float = 1.0
    asm_version: str = ""
    asm_end_marker: bool = False  # depends on DXF version: R2013, RT2018

    @property
    def has_asm_header(self) -> bool:
        return self.asm_version != ""

    def dumps(self) -> list[str]:
        """Returns the SAT file header as list of strings."""
        return [
            f"{self.version} {self.n_records} {self.n_entities} {self.flags} ",
            self._header_str(),
            f"{self.units_in_mm:g} 9.9999999999999995e-007 1e-010 ",
        ]

    def dumpb(self) -> bytes:
        """Returns the SAB file header as bytes."""
        buffer: list[bytes] = []
        if self.version > 21800:
            buffer.append(const.ASM_SIGNATURE)
        else:
            buffer.append(const.ACIS_SIGNATURE)
        data = struct.pack(
            "<iiii", self.version, self.n_records, self.n_entities, self.flags
        )
        buffer.append(data)
        buffer.append(encode_str(self.product_id))
        buffer.append(encode_str(self.acis_version))
        buffer.append(encode_str(self.creation_date.ctime()))
        buffer.append(encode_double(self.units_in_mm))
        buffer.append(encode_double(const.RES_TOL))
        buffer.append(encode_double(const.NOR_TOL))
        return b"".join(buffer)

    def _header_str(self) -> str:
        p_len = len(self.product_id)
        a_len = len(self.acis_version)
        date = self.creation_date.ctime()
        if self.version > 400:
            return f"@{p_len} {self.product_id} @{a_len} {self.acis_version} @{len(date)} {date} "
        else:
            return f"{p_len} {self.product_id} {a_len} {self.acis_version} {len(date)} {date} "

    def set_version(self, version: int) -> None:
        """Sets the ACIS version as an integer value and updates the version
        string accordingly.
        """
        try:
            self.acis_version = const.ACIS_VERSION[version]
            self.version = version
        except KeyError:
            raise ValueError(f"invalid ACIS version number {version}")
        self.asm_version = const.ASM_VERSION.get(version, "")

    def asm_header(self):
        from .entities import AsmHeader
        return AsmHeader(self.asm_version)

    def sat_end_marker(self) -> str:
        if self.asm_end_marker:
            return const.END_OF_ASM_DATA_SAT + " "
        else:
            return const.END_OF_ACIS_DATA_SAT + " "

    def sab_end_marker(self) -> bytes:
        if self.asm_end_marker:
            return const.END_OF_ASM_DATA_SAB
        else:
            return const.END_OF_ACIS_DATA_SAB



# Copyright (c) 2024 Manfred Moitzi
# License: MIT License
import pytest
import ezdxf
from ezdxf.layouts import Modelspace


@pytest.fixture(scope="module")
def msp() -> Modelspace:
    doc = ezdxf.new("R2007", setup=True)
    return doc.modelspace()


@pytest.mark.parametrize("text", ["", "<>"])
def test_regular_measurement_text(msp, text):
    dim = msp.add_linear_dim(
        base=(0, 2),
        p1=(0, 0),
        p2=(3, 0),
    )
    dim.set_text(text)

    block = dim.render().geometry.layout
    mtext = block.query("MTEXT").first
    assert mtext.text == "300"


def test_suppress_measurement_text(msp):
    dim = msp.add_linear_dim(
        base=(0, 2),
        p1=(0, 0),
        p2=(3, 0),
    )
    dim.set_text(" ")

    block = dim.render().geometry.layout
    assert len(block.query("MTEXT")) == 0  # no MTEXT entity expected


@pytest.mark.parametrize(
    "text,expected",
    [
        ("<>", "300"),
        ("Ø <>", "Ø 300"),
        ("<> cm", "300 cm"),
        ("[ <> ]", "[ 300 ]"),
        ("<> <>", "300 <>"),  # only the first "<>" will be replaced
    ],
)
def test_override_measurement_text(msp, text, expected):
    dim = msp.add_linear_dim(
        base=(0, 2),
        p1=(0, 0),
        p2=(3, 0),
    )
    dim.set_text(text)

    block = dim.render().geometry.layout
    mtext = block.query("MTEXT").first
    assert mtext.text == expected


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License

import pytest
import math
from ezdxf.math.bspline import bspline_basis_vector, open_uniform_knot_vector
from ezdxf.math._bspline import Basis
from ezdxf.acc import USE_C_EXT

basis_functions = [Basis]

if USE_C_EXT:
    pass


@pytest.fixture(params=basis_functions)
def basis_cls(request):
    return request.param


def make_knots(count, degree):
    return tuple(open_uniform_knot_vector(count, order=degree + 1))


def make_basic_func(count, degree, cls):
    knots = make_knots(count, degree)
    return cls(knots=knots, order=degree + 1, count=count)


def test_property_exists(basis_cls):
    degree = 3
    count = 10
    knots = make_knots(count, degree)
    basis_func = basis_cls(knots=knots, order=degree + 1, count=count)
    assert basis_func.degree == degree
    assert basis_func.order == degree + 1
    assert basis_func.max_t == max(knots)
    assert basis_func.knots == knots
    assert basis_func.weights == tuple()
    assert basis_func.is_rational is False


def test_bspline_basis_vector(basis_cls):
    degree = 3
    count = 10
    knots = make_knots(count, degree)
    max_t = max(knots)
    basis_func = basis_cls(knots=knots, order=degree + 1, count=count)
    for u in (0, 2, 2.5, 3.5, 4, max_t):
        basis = bspline_basis_vector(u, count=count, degree=degree, knots=knots)
        basis2 = basis_func.basis_vector(u)
        assert len(basis) == len(basis2)
        for v1, v2 in zip(basis, basis2):
            assert v1 == pytest.approx(v2)


def test_find_span(basis_cls):
    degree = 3
    basis_func = make_basic_func(10, degree, basis_cls)
    for u in [0, 2, 2.5, 3.5, 4]:
        result = basis_func.find_span(u)
        span = math.floor(u) + degree
        assert result == span
    assert basis_func.find_span(basis_func.max_t) == 9


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2019-2023 Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import TYPE_CHECKING, Union, Optional
from ezdxf.lldxf.tags import Tags
from ezdxf.lldxf.const import DXFStructureError
from ezdxf.lldxf.const import (
    ACAD_XDICTIONARY,
    XDICT_HANDLE_CODE,
    APP_DATA_MARKER,
)
from .copy import default_copy

if TYPE_CHECKING:
    from ezdxf.document import Drawing
    from ezdxf.lldxf.tagwriter import AbstractTagWriter
    from ezdxf.entities import (
        Dictionary,
        DXFEntity,
        DXFObject,
        Placeholder,
        DictionaryVar,
        XRecord,
    )

__all__ = ["ExtensionDict"]


# Example for table head and -entries with extension dicts:
# AutodeskSamples\lineweights.dxf


class ExtensionDict:
    """Stores extended data of entities in app data 'ACAD_XDICTIONARY', app
    data contains just one entry to a hard-owned DICTIONARY objects, which is
    not shared with other entities, each entity copy has its own extension
    dictionary and the extension dictionary is destroyed when the owner entity
    is deleted from database.

    """

    __slots__ = ("_xdict",)

    def __init__(self, xdict: Union[str, Dictionary]):
        # 1st loading stage: xdict as string -> handle to dict
        # 2nd loading stage: xdict as DXF Dictionary
        self._xdict = xdict

    @property
    def dictionary(self) -> Dictionary:
        """Returns the underlying :class:`~ezdxf.entities.Dictionary` object."""
        xdict = self._xdict
        assert xdict is not None, "destroyed extension dictionary"
        assert not isinstance(xdict, str), f"dictionary handle #{xdict} not resolved"
        return xdict

    @property
    def handle(self) -> str:
        """Returns the handle of the underlying :class:`~ezdxf.entities.Dictionary`
        object.
        """
        return self.dictionary.dxf.handle

    def __getitem__(self, key: str):
        """Get self[key]."""
        return self.dictionary[key]

    def __setitem__(self, key: str, value):
        """Set self[key] to value.

        Only DXF objects stored in the OBJECTS section are allowed as content
        of the extension dictionary. DXF entities stored in layouts are not
        allowed.

        Raises:
            DXFTypeError: invalid DXF type

        """
        self.dictionary[key] = value

    def __delitem__(self, key: str):
        """Delete self[key], destroys referenced entity."""
        del self.dictionary[key]

    def __contains__(self, key: str):
        """Return `key` in self."""
        return key in self.dictionary

    def __len__(self):
        """Returns count of extension dictionary entries."""
        return len(self.dictionary)

    def keys(self):
        """Returns a :class:`KeysView` of all extension dictionary keys."""
        return self.dictionary.keys()

    def items(self):
        """Returns an :class:`ItemsView` for all extension dictionary entries as
        (key, entity) pairs. An entity can be a handle string if the entity
        does not exist.
        """
        return self.dictionary.items()

    def get(self, key: str, default=None) -> Optional[DXFEntity]:
        """Return extension dictionary entry `key`."""
        return self.dictionary.get(key, default)

    def discard(self, key: str) -> None:
        """Discard extension dictionary entry `key`."""
        return self.dictionary.discard(key)

    @classmethod
    def new(cls, owner_handle: str, doc: Drawing):
        xdict = doc.objects.add_dictionary(
            owner=owner_handle,
            # All data in the extension dictionary belongs only to the owner
            hard_owned=True,
        )
        return cls(xdict)

    def copy(self, copy_strategy=default_copy) -> ExtensionDict:
        """Deep copy of the extension dictionary all entries are virtual
        entities.
        """
        new_xdict = copy_strategy.copy(self.dictionary)
        return ExtensionDict(new_xdict)

    @property
    def is_alive(self):
        """Returns ``True`` if the underlying :class:`~ezdxf.entities.Dictionary`
        object is not deleted.
        """
        # Can not check if _xdict (as handle or Dictionary) really exist:
        return self._xdict is not None

    @property
    def has_valid_dictionary(self):
        """Returns ``True`` if the underlying :class:`~ezdxf.entities.Dictionary`
        really exist and is valid.
        """
        xdict = self._xdict
        if xdict is None or isinstance(xdict, str):
            return False
        return xdict.is_alive

    def update_owner(self, handle: str) -> None:
        """Update owner tag of underlying :class:`~ezdxf.entities.Dictionary`
        object.

        Internal API.
        """
        assert self.is_alive, "destroyed extension dictionary"
        self.dictionary.dxf.owner = handle

    @classmethod
    def from_tags(cls, tags: Tags):
        assert tags is not None
        # Expected DXF structure:
        # [(102, '{ACAD_XDICTIONARY', (360, handle), (102, '}')]
        if len(tags) != 3 or tags[1].code != XDICT_HANDLE_CODE:
            raise DXFStructureError("ACAD_XDICTIONARY error.")
        return cls(tags[1].value)

    def load_resources(self, doc: Drawing) -> None:
        handle = self._xdict
        assert isinstance(handle, str)
        self._xdict = doc.entitydb.get(handle)  # type: ignore

    def export_dxf(self, tagwriter: AbstractTagWriter) -> None:
        assert self._xdict is not None
        xdict = self._xdict
        handle = xdict if isinstance(xdict, str) else xdict.dxf.handle
        tagwriter.write_tag2(APP_DATA_MARKER, ACAD_XDICTIONARY)
        tagwriter.write_tag2(XDICT_HANDLE_CODE, handle)
        tagwriter.write_tag2(APP_DATA_MARKER, "}")

    def destroy(self):
        """Destroy the underlying :class:`~ezdxf.entities.Dictionary` object."""
        if self.has_valid_dictionary:
            self._xdict.destroy()
        self._xdict = None

    def add_dictionary(self, name: str, hard_owned: bool = True) -> Dictionary:
        """Create a new :class:`~ezdxf.entities.Dictionary` object as
        extension dictionary entry `name`.
        """
        dictionary = self.dictionary
        doc = dictionary.doc
        assert doc is not None, "valid DXF document required"
        new_dict = doc.objects.add_dictionary(
            owner=dictionary.dxf.handle,
            hard_owned=hard_owned,
        )
        dictionary[name] = new_dict
        return new_dict

    def add_xrecord(self, name: str) -> XRecord:
        """Create a new :class:`~ezdxf.entities.XRecord` object as
        extension dictionary entry `name`.
        """
        dictionary = self.dictionary
        doc = dictionary.doc
        assert doc is not None, "valid DXF document required"
        xrecord = doc.objects.add_xrecord(dictionary.dxf.handle)
        dictionary[name] = xrecord
        return xrecord

    def add_dictionary_var(self, name: str, value: str) -> DictionaryVar:
        """Create a new :class:`~ezdxf.entities.DictionaryVar` object as
        extension dictionary entry `name`.
        """
        dictionary = self.dictionary
        doc = dictionary.doc
        assert doc is not None, "valid DXF document required"
        dict_var = doc.objects.add_dictionary_var(dictionary.dxf.handle, value)
        dictionary[name] = dict_var
        return dict_var

    def add_placeholder(self, name: str) -> Placeholder:
        """Create a new :class:`~ezdxf.entities.Placeholder` object as
        extension dictionary entry `name`.
        """
        dictionary = self.dictionary
        doc = dictionary.doc
        assert doc is not None, "valid DXF document required"
        placeholder = doc.objects.add_placeholder(dictionary.dxf.handle)
        dictionary[name] = placeholder
        return placeholder

    def link_dxf_object(self, name: str, obj: DXFObject) -> None:
        """Link `obj` to the extension dictionary as entry `name`.

        Linked objects are owned by the extensions dictionary and therefore
        cannot be a graphical entity, which have to be owned by a
        :class:`~ezdxf.layouts.BaseLayout`.

        Raises:
            DXFTypeError: `obj` has invalid DXF type

        """
        self.dictionary.link_dxf_object(name, obj)



# Copyright (c) 2020-2021, Manfred Moitzi
# License: MIT License
from typing import Dict, Tuple
import struct
from .const import *
from .crc import crc8

codepage_to_encoding = {
    37: "cp874",  # Thai,
    38: "cp932",  # Japanese
    39: "gbk",  # UnifiedChinese
    40: "cp949",  # Korean
    41: "cp950",  # TradChinese
    28: "cp1250",  # CentralEurope
    29: "cp1251",  # Cyrillic
    30: "cp1252",  # WesternEurope
    32: "cp1253",  # Greek
    33: "cp1254",  # Turkish
    34: "cp1255",  # Hebrew
    35: "cp1256",  # Arabic
    36: "cp1257",  # Baltic
}

FILE_HEADER_MAGIC = {
    3: 0xA598,
    4: 0x8101,
    5: 0x3CC4,
    6: 0x8461,
}


class FileHeader:
    def __init__(self, data: Bytes, crc_check=False):
        self.crc_check = crc_check
        if len(data) < 6:
            raise DwgVersionError("Not a DWG file.")
        ver = data[:6].decode(errors="ignore")  # type: ignore
        if ver not in SUPPORTED_VERSIONS:
            raise DwgVersionError(
                f"Not a DWG file or unsupported DWG version, signature: {ver}."
            )
        self.version: str = ver
        codepage: int = struct.unpack_from("<h", data, 0x13)[0]
        self.encoding = codepage_to_encoding.get(codepage, "cp1252")
        self.maintenance_release_version = data[0xB]
        self.sections: Dict[int, Tuple[int, int]] = dict()
        if self.version <= ACAD_2000:
            self.r2000_header(data)
        else:
            raise DwgVersionError(self.version)

    def r2000_header(self, data: Bytes):
        index = 0x15
        section_count: int = struct.unpack_from("<L", data, index)[0]
        index += 4
        fmt = "<BLL"
        record_size = struct.calcsize(fmt)
        for record in range(section_count):
            # 0: HEADER_ID
            # 1: CLASSES_ID
            # 2: OBJECTS_ID
            num, seeker, size = struct.unpack_from(fmt, data, index)
            index += record_size
            self.sections[num] = (seeker, size)

        if self.crc_check:
            # CRC from first byte of file until start of crc value
            check = (
                crc8(data[:index], seed=0)
                ^ FILE_HEADER_MAGIC[len(self.sections)]
            )
            crc = struct.unpack_from("<H", data, index)[0]
            if crc != check:
                raise CRCError("CRC error in file header.")

        index += 2
        sentinel = data[index : index + SENTINEL_SIZE]
        if (
            sentinel
            != b"\x95\xA0\x4E\x28\x99\x82\x1A\xE5\x5E\x41\xE0\x5F\x9D\x3A\x4D\x00"
        ):
            raise DwgCorruptedFileHeader(
                "Corrupted DXF R13/14/2000 file header."
            )

    def print(self):
        print(f"DWG version: {self.version}")
        print(f"encoding: {self.encoding}")
        print(f"Records: {len(self.sections)}")
        print("Header: seeker {0[0]} size: {0[1]}".format(self.sections[0]))
        print("Classes: seeker {0[0]} size: {0[1]}".format(self.sections[1]))
        print("Objects: seeker {0[0]} size: {0[1]}".format(self.sections[2]))



# Copyright (c) 2024, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import Sequence, TYPE_CHECKING, Iterable, Any
import math
from ezdxf.math import UVec, Vec2, has_clockwise_orientation
from ezdxf.entities import LWPolyline, DXFEntity

if TYPE_CHECKING:
    from ezdxf.layouts import BaseLayout


REVCLOUD_PROPS = "RevcloudProps"
REQUIRED_BULGE = 0.520567050552


def points(
    vertices: Iterable[UVec],
    segment_length: float,
    *,
    bulge: float = REQUIRED_BULGE,  # required to be recognized as REVCLOUD
    start_width: float = 0.0,
    end_width: float = 0.0,
) -> list[Sequence[float]]:
    """Returns the points for a :class:`~ezdxf.entities.LWPolyline` entity to render a
    revision cloud, similar to the REVCLOUD command in CAD applications.

    Args:
        vertices: corner points of a polygon
        segment_length: approximate segment length
        bulge: LWPOLYLINE bulge value
        start_width: start width of the segment arc
        end_width: end width of the segment arc, CAD applications use 0.1 * segment_length
            for a calligraphy effect

    """
    if segment_length < 1e-6:
        raise ValueError("segment length too small.")
    _vertices = _to_vec2_list(vertices)
    lw_points: list[Sequence[float]] = []

    for s, e in zip(_vertices, _vertices[1:]):
        lw_points.append((s.x, s.y, start_width, end_width, bulge))
        diff = e - s
        length = diff.magnitude
        if length <= segment_length:
            continue

        count = math.ceil(length / segment_length)
        _segment_length = length / count
        offset = diff.normalize(_segment_length)
        for _ in range(count - 1):
            s += offset
            lw_points.append((s.x, s.y, start_width, end_width, bulge))
    return lw_points


def add_entity(
    layout: BaseLayout,
    vertices: Iterable[UVec],
    segment_length: float,
    *,
    calligraphy=True,
    dxfattribs: Any = None,
) -> LWPolyline:
    """Adds a revision cloud as :class:`~ezdxf.entities.LWPolyline` entity to `layout`,
    similar to the REVCLOUD command in CAD applications.

    Args:
        layout: target layout
        vertices: corner points of a polygon
        segment_length: approximate segment length
        calligraphy: ``True`` for a calligraphy effect
        dxfattribs: additional DXF attributes

    """
    _vertices = _to_vec2_list(vertices)
    bulge = REQUIRED_BULGE
    if has_clockwise_orientation(_vertices):
        bulge = -bulge

    end_width = segment_length * 0.1 if calligraphy else 0.0
    lw_points = points(_vertices, segment_length, bulge=bulge, end_width=end_width)
    lwp = layout.add_lwpolyline(lw_points, close=True, dxfattribs=dxfattribs)
    lwp.set_xdata(REVCLOUD_PROPS, [(1070, 0), (1040, segment_length)])
    doc = layout.doc
    if doc is not None and not doc.appids.has_entry(REVCLOUD_PROPS):
        doc.appids.add(REVCLOUD_PROPS)
    return lwp


def is_revcloud(entity: DXFEntity) -> bool:
    """Returns ``True`` when the given entity represents a revision cloud."""
    if not isinstance(entity, LWPolyline):
        return False
    lwpolyline: LWPolyline = entity
    if not lwpolyline.is_closed:
        return False
    if not lwpolyline.has_xdata(REVCLOUD_PROPS):
        return False
    return all(
        abs(REQUIRED_BULGE - abs(p[0])) < 0.02
        for p in lwpolyline.get_points(format="b")
    )


def _to_vec2_list(vertices: Iterable[UVec]) -> list[Vec2]:
    _vertices: list[Vec2] = Vec2.list(vertices)
    if len(_vertices) < 3:
        raise ValueError("3 or more points required.")
    if not _vertices[0].isclose(_vertices[-1]):
        _vertices.append(_vertices[0])
    if len(_vertices) < 4:
        raise ValueError("3 or more points required.")
    return _vertices



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License

import pytest
from io import StringIO
from time import sleep
import ezdxf


def doc2str(doc):
    stream = StringIO()
    doc.write(stream)
    return stream.getvalue()


@pytest.mark.parametrize("version", ["R12", "R2000", "R2018"])
def test_fixed_meta_data(version):
    ezdxf.options.write_fixed_meta_data_for_testing = True
    doc = ezdxf.new(version)
    txt1 = doc2str(doc)
    # write it again
    sleep(0.1)
    txt2 = doc2str(doc)
    assert txt1 == txt2


@pytest.mark.parametrize("version", ["R12", "R2000", "R2018"])
def test_meta_data_is_different(version):
    ezdxf.options.write_fixed_meta_data_for_testing = False
    doc = ezdxf.new(version)
    txt1 = doc2str(doc)
    # write it again
    sleep(0.1)
    txt2 = doc2str(doc)
    assert txt1 != txt2


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2019-2024, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import TYPE_CHECKING, Optional, cast, Any
from typing_extensions import Self
import logging
from dataclasses import dataclass
from ezdxf.lldxf import validator
from ezdxf.lldxf.attributes import (
    DXFAttr,
    DXFAttributes,
    DefSubclass,
    RETURN_DEFAULT,
    group_code_mapping,
)
from ezdxf import colors as clr
from ezdxf.lldxf import const
from ezdxf.lldxf.const import (
    DXF12,
    SUBCLASS_MARKER,
    DXF2000,
    DXF2007,
    DXF2004,
    INVALID_NAME_CHARACTERS,
    DXFValueError,
    LINEWEIGHT_BYBLOCK,
    LINEWEIGHT_BYLAYER,
    LINEWEIGHT_DEFAULT,
)
from ezdxf.audit import AuditError
from ezdxf.entities.dxfentity import base_class, SubclassProcessor, DXFEntity
from .factory import register_entity

if TYPE_CHECKING:
    from ezdxf.entities import DXFNamespace, Viewport, XRecord
    from ezdxf.lldxf.tagwriter import AbstractTagWriter
    from ezdxf.entitydb import EntityDB
    from ezdxf import xref
    from ezdxf.audit import Auditor


__all__ = ["Layer", "acdb_symbol_table_record", "LayerOverrides"]
logger = logging.getLogger("ezdxf")


def is_valid_layer_color_index(aci: int) -> bool:
    # BYBLOCK or BYLAYER is not valid a layer color!
    return (-256 < aci < 256) and aci != 0


def fix_layer_color(aci: int) -> int:
    return aci if is_valid_layer_color_index(aci) else 7


def is_valid_layer_lineweight(lw: int) -> bool:
    if validator.is_valid_lineweight(lw):
        if lw not in (LINEWEIGHT_BYLAYER, LINEWEIGHT_BYBLOCK):
            return True
    return False


def fix_layer_lineweight(lw: int) -> int:
    if lw in (LINEWEIGHT_BYLAYER, LINEWEIGHT_BYBLOCK):
        return LINEWEIGHT_DEFAULT
    else:
        return validator.fix_lineweight(lw)


acdb_symbol_table_record: DefSubclass = DefSubclass("AcDbSymbolTableRecord", {})

acdb_layer_table_record = DefSubclass(
    "AcDbLayerTableRecord",
    {
        # Layer name as string
        "name": DXFAttr(2, validator=validator.is_valid_layer_name),
        "flags": DXFAttr(70, default=0),
        # ACI color index, color < 0 indicates layer status: off
        "color": DXFAttr(
            62,
            default=7,
            validator=is_valid_layer_color_index,
            fixer=fix_layer_color,
        ),
        # True color as 24 bit int value: 0x00RRGGBB
        "true_color": DXFAttr(420, dxfversion=DXF2004, optional=True),
        # Linetype name as string
        "linetype": DXFAttr(
            6, default="Continuous", validator=validator.is_valid_table_name
        ),
        # 0 = don't plot layer; 1 = plot layer
        "plot": DXFAttr(
            290,
            default=1,
            dxfversion=DXF2000,
            optional=True,
            validator=validator.is_integer_bool,
            fixer=RETURN_DEFAULT,
        ),
        # Default lineweight 1/100 mm, min 0 = 0.0mm, max 211 = 2.11mm
        "lineweight": DXFAttr(
            370,
            default=LINEWEIGHT_DEFAULT,
            dxfversion=DXF2000,
            validator=is_valid_layer_lineweight,
            fixer=fix_layer_lineweight,
        ),
        # Handle to PlotStyleName, group code 390 is required by AutoCAD
        "plotstyle_handle": DXFAttr(390, dxfversion=DXF2000),
        # Handle to Material object
        "material_handle": DXFAttr(347, dxfversion=DXF2007),
        # Handle to ???
        "unknown1": DXFAttr(348, dxfversion=DXF2007, optional=True),
    },
)
acdb_layer_table_record_group_codes = group_code_mapping(acdb_layer_table_record)
AcAecLayerStandard = "AcAecLayerStandard"
AcCmTransparency = "AcCmTransparency"


@register_entity
class Layer(DXFEntity):
    """DXF LAYER entity"""

    DXFTYPE = "LAYER"
    DXFATTRIBS = DXFAttributes(
        base_class, acdb_symbol_table_record, acdb_layer_table_record
    )
    DEFAULT_ATTRIBS = {"name": "0"}
    FROZEN = 0b00000001
    THAW = 0b11111110
    LOCK = 0b00000100
    UNLOCK = 0b11111011

    def load_dxf_attribs(
        self, processor: Optional[SubclassProcessor] = None
    ) -> DXFNamespace:
        dxf = super().load_dxf_attribs(processor)
        if processor:
            processor.simple_dxfattribs_loader(
                dxf, acdb_layer_table_record_group_codes  # type: ignore
            )
        return dxf

    def export_entity(self, tagwriter: AbstractTagWriter) -> None:
        super().export_entity(tagwriter)
        if tagwriter.dxfversion > DXF12:
            tagwriter.write_tag2(SUBCLASS_MARKER, acdb_symbol_table_record.name)
            tagwriter.write_tag2(SUBCLASS_MARKER, acdb_layer_table_record.name)

        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                "name",
                "flags",
                "color",
                "true_color",
                "linetype",
                "plot",
                "lineweight",
                "plotstyle_handle",
                "material_handle",
                "unknown1",
            ],
        )

    def set_required_attributes(self):
        assert self.doc is not None, "valid DXF document required"
        if not self.dxf.hasattr("material_handle"):
            global_ = self.doc.materials["Global"]
            if isinstance(global_, DXFEntity):
                handle = global_.dxf.handle
            else:
                handle = global_
            self.dxf.material_handle = handle
        if not self.dxf.hasattr("plotstyle_handle"):
            normal = self.doc.plotstyles["Normal"]
            if isinstance(normal, DXFEntity):
                handle = normal.dxf.handle
            else:
                handle = normal
            self.dxf.plotstyle_handle = handle

    def is_frozen(self) -> bool:
        """Returns ``True`` if layer is frozen."""
        return self.dxf.flags & Layer.FROZEN > 0

    def freeze(self) -> None:
        """Freeze layer."""
        self.dxf.flags = self.dxf.flags | Layer.FROZEN

    def thaw(self) -> None:
        """Thaw layer."""
        self.dxf.flags = self.dxf.flags & Layer.THAW

    def is_locked(self) -> bool:
        """Returns ``True`` if layer is locked."""
        return self.dxf.flags & Layer.LOCK > 0

    def lock(self) -> None:
        """Lock layer, entities on this layer are not editable - just important
        in CAD applications.
        """
        self.dxf.flags = self.dxf.flags | Layer.LOCK

    def unlock(self) -> None:
        """Unlock layer, entities on this layer are editable - just important
        in CAD applications.
        """
        self.dxf.flags = self.dxf.flags & Layer.UNLOCK

    def is_off(self) -> bool:
        """Returns ``True`` if layer is off."""
        return self.dxf.color < 0

    def is_on(self) -> bool:
        """Returns ``True`` if layer is on."""
        return not self.is_off()

    def on(self) -> None:
        """Switch layer `on` (visible)."""
        self.dxf.color = abs(self.dxf.color)

    def off(self) -> None:
        """Switch layer `off` (invisible)."""
        self.dxf.color = -abs(self.dxf.color)

    def get_color(self) -> int:
        """Get layer color, safe method for getting the layer color, because
        :attr:`dxf.color` is negative for layer status `off`.
        """
        return abs(self.dxf.color)

    def set_color(self, color: int) -> None:
        """Set layer color, safe method for setting the layer color, because
        :attr:`dxf.color` is negative for layer status `off`.
        """
        color = abs(color) if self.is_on() else -abs(color)
        self.dxf.color = color

    @property
    def rgb(self) -> Optional[tuple[int, int, int]]:
        """Returns RGB true color as (r, g, b)-tuple or None if attribute
        dxf.true_color is not set.
        """
        if self.dxf.hasattr("true_color"):
            return clr.int2rgb(self.dxf.get("true_color"))
        else:
            return None

    @rgb.setter
    def rgb(self, rgb: tuple[int, int, int]) -> None:
        """Set RGB true color as (r, g, b)-tuple e.g. (12, 34, 56)."""
        self.dxf.set("true_color", clr.rgb2int(rgb))

    @property
    def color(self) -> int:
        """Get layer color, safe method for getting the layer color, because
        :attr:`dxf.color` is negative for layer status `off`.
        """
        return self.get_color()

    @color.setter
    def color(self, value: int) -> None:
        """Set layer color, safe method for setting the layer color, because
        :attr:`dxf.color` is negative for layer status `off`.
        """
        self.set_color(value)

    @property
    def description(self) -> str:
        try:
            xdata = self.get_xdata(AcAecLayerStandard)
        except DXFValueError:
            return ""
        else:
            if len(xdata) > 1:
                # this is the usual case in BricsCAD
                return xdata[1].value
            else:
                return ""

    @description.setter
    def description(self, value: str) -> None:
        # create AppID table entry if not present
        if self.doc and AcAecLayerStandard not in self.doc.appids:
            self.doc.appids.new(AcAecLayerStandard)
        self.discard_xdata(AcAecLayerStandard)
        self.set_xdata(AcAecLayerStandard, [(1000, ""), (1000, value)])

    @property
    def transparency(self) -> float:
        try:
            xdata = self.get_xdata(AcCmTransparency)
        except DXFValueError:
            return 0.0
        else:
            t = xdata[0].value
            if t & 0x2000000:  # is this a real transparency value?
                # Transparency BYBLOCK (0x01000000) make no sense for a layer!?
                return clr.transparency2float(t)
        return 0.0

    @transparency.setter
    def transparency(self, value: float) -> None:
        # create AppID table entry if not present
        if self.doc and AcCmTransparency not in self.doc.appids:
            self.doc.appids.new(AcCmTransparency)
        if 0 <= value <= 1:
            self.discard_xdata(AcCmTransparency)
            self.set_xdata(AcCmTransparency, [(1071, clr.float2transparency(value))])
        else:
            raise ValueError("Value out of range [0, 1].")

    def rename(self, name: str) -> None:
        """
        Rename layer and all known (documented) references to this layer.

        .. warning::

            The DXF format is not consistent in storing layer references, the
            layers are mostly referenced by their case-insensitive name,
            some later introduced entities do reference layers by handle, which
            is the safer way in the context of renaming layers.

            There is no complete overview of where layer references are
            stored, third-party entities are black-boxes with unknown content
            and layer names could be stored in the extended data section of any
            DXF entity or in XRECORD entities.
            Which means that in some rare cases references to the old layer name
            can persist, at least this does not invalidate the DXF document.

        Args:
             name: new layer name

        Raises:
            ValueError: `name` contains invalid characters: <>/\\":;?*|=`
            ValueError: layer `name` already exist
            ValueError: renaming of layers ``'0'`` and ``'DEFPOINTS'`` not
                possible

        """
        if not validator.is_valid_layer_name(name):
            raise ValueError(
                f"Name contains invalid characters: {INVALID_NAME_CHARACTERS}."
            )
        assert self.doc is not None, "valid DXF document is required"
        layers = self.doc.layers
        if self.dxf.name.lower() in ("0", "defpoints"):
            raise ValueError(f'Can not rename layer "{self.dxf.name}".')
        if layers.has_entry(name):
            raise ValueError(f'Layer "{name}" already exist.')
        old = self.dxf.name
        self.dxf.name = name
        layers.replace(old, self)
        self._rename_layer_references(old, name)

    def _rename_layer_references(self, old_name: str, new_name: str) -> None:
        assert self.doc is not None, "valid DXF document is required"
        key = self.doc.layers.key
        old_key = key(old_name)
        for e in self.doc.entitydb.values():
            if e.dxf.hasattr("layer") and key(e.dxf.layer) == old_key:
                e.dxf.layer = new_name
            entity_type = e.dxftype()
            if entity_type == "VIEWPORT":
                e.rename_frozen_layer(old_name, new_name)  # type: ignore
            elif entity_type == "LAYER_FILTER":
                # todo: if LAYER_FILTER implemented, add support for
                #  renaming layers
                logger.debug(
                    f'renaming layer "{old_name}" - document contains ' f"LAYER_FILTER"
                )
            elif entity_type == "LAYER_INDEX":
                # todo: if LAYER_INDEX implemented, add support for
                #  renaming layers
                logger.debug(
                    f'renaming layer "{old_name}" - document contains ' f"LAYER_INDEX"
                )

    def get_vp_overrides(self) -> LayerOverrides:
        """Returns the :class:`LayerOverrides` object for this layer."""
        return LayerOverrides(self)

    def register_resources(self, registry: xref.Registry) -> None:
        """Register required resources to the resource registry."""
        assert self.doc is not None, "LAYER entity must be assigned to a document"
        super().register_resources(registry)
        registry.add_linetype(self.dxf.linetype)
        registry.add_handle(self.dxf.get("material_handle"))
        # current plot style will be replaced by default plot style "Normal"

    def map_resources(self, clone: Self, mapping: xref.ResourceMapper) -> None:
        """Translate resources from self to the copied entity."""
        assert isinstance(clone, Layer)
        super().map_resources(clone, mapping)
        self.dxf.linetype = mapping.get_linetype(self.dxf.linetype)

        mapping.map_existing_handle(self, clone, "material_handle", optional=True)
        # remove handles pointing to the source document:
        clone.dxf.discard("plotstyle_handle")  # replaced by plot style "Normal"
        clone.dxf.discard("unknown1")

        # create required handles to resources in the target document
        clone.set_required_attributes()
        # todo: map layer overrides
        # remove layer overrides
        clone.discard_extension_dict()

    def audit(self, auditor: Auditor) -> None:
        super().audit(auditor)
        linetype = self.dxf.linetype
        if auditor.doc.linetypes.has_entry(linetype):
            return
        self.dxf.linetype = "Continuous"
        auditor.fixed_error(
            code=AuditError.UNDEFINED_LINETYPE,
            message=f"Replaced undefined linetype {linetype} in layer '{self.dxf.name}' by CONTINUOUS",
            dxf_entity=self,
            data=linetype,
        )


@dataclass
class OverrideAttributes:
    aci: int
    rgb: Optional[clr.RGB]
    transparency: float
    linetype: str
    lineweight: int


class LayerOverrides:
    """This object stores the layer attribute overridden in VIEWPORT entities,
    where each VIEWPORT can have individual layer attribute overrides.

    Layer attributes which can be overridden:

        - ACI color
        - true color (rgb)
        - linetype
        - lineweight
        - transparency

    """

    def __init__(self, layer: Layer):
        assert layer.doc is not None, "valid DXF document required"
        self._layer = layer
        self._overrides = load_layer_overrides(layer)

    def has_overrides(self, vp_handle: Optional[str] = None) -> bool:
        """Returns ``True`` if attribute overrides exist for the given
        :class:`Viewport` handle.
        Returns ``True`` if `any` attribute overrides exist if the given
        handle is ``None``.
        """
        if vp_handle is None:
            return bool(self._overrides)
        return vp_handle in self._overrides

    def commit(self) -> None:
        """Write :class:`Viewport` overrides back into the :class:`Layer` entity.
        Without a commit() all changes are lost!
        """
        store_layer_overrides(self._layer, self._overrides)

    def _acquire_overrides(self, vp_handle: str) -> OverrideAttributes:
        """Returns the OverrideAttributes() instance for `vp_handle`, creates a new
        OverrideAttributes() instance if none exist.
        """
        return self._overrides.setdefault(
            vp_handle,
            default_ovr_settings(self._layer),
        )

    def _get_overrides(self, vp_handle: str) -> OverrideAttributes:
        """Returns the overrides for `vp_handle`, returns the default layer
        settings if no Override() instance exist.
        """
        try:
            return self._overrides[vp_handle]
        except KeyError:
            return default_ovr_settings(self._layer)

    def set_color(self, vp_handle: str, value: int) -> None:
        """Override the :ref:`ACI`.

        Raises:
            ValueError: invalid color value
        """
        # BYBLOCK or BYLAYER is not valid a layer color
        if not is_valid_layer_color_index(value):
            raise ValueError(f"invalid ACI value: {value}")
        vp_overrides = self._acquire_overrides(vp_handle)
        vp_overrides.aci = value

    def get_color(self, vp_handle: str) -> int:
        """Returns the :ref:`ACI` override or the original layer value if no
        override exist.
        """
        vp_overrides = self._get_overrides(vp_handle)
        return vp_overrides.aci

    def set_rgb(self, vp_handle: str, value: Optional[clr.RGB]):
        """Set the RGB override as (red, gree, blue) tuple or ``None`` to remove
        the true color setting.

        Raises:
            ValueError: invalid RGB value

        """
        if value is not None and not validator.is_valid_rgb(value):
            raise ValueError(f"invalid RGB value: {value}")
        vp_overrides = self._acquire_overrides(vp_handle)
        vp_overrides.rgb = value

    def get_rgb(self, vp_handle: str) -> Optional[clr.RGB]:
        """Returns the RGB override or the original layer value if no
        override exist. Returns ``None`` if no true color value is set.
        """
        vp_overrides = self._get_overrides(vp_handle)
        return vp_overrides.rgb

    def set_transparency(self, vp_handle: str, value: float) -> None:
        """Set the transparency override. A transparency of 0.0 is opaque and
        1.0 is fully transparent.

        Raises:
            ValueError: invalid transparency value

        """
        if not (0.0 <= value <= 1.0):
            raise ValueError(
                f"invalid transparency: {value}, has to be in the range [0, 1]"
            )
        vp_overrides = self._acquire_overrides(vp_handle)
        vp_overrides.transparency = value

    def get_transparency(self, vp_handle: str) -> float:
        """Returns the transparency override or the original layer value if no
        override exist. Returns 0.0 for opaque and 1.0 for fully transparent.
        """
        vp_overrides = self._get_overrides(vp_handle)
        return vp_overrides.transparency

    def set_linetype(self, vp_handle: str, value: str) -> None:
        """Set the linetype override.

        Raises:
            ValueError: linetype without a LTYPE table entry
        """
        if value not in self._layer.doc.linetypes:  # type: ignore
            raise ValueError(
                f"invalid linetype: {value}, a linetype table entry is required"
            )
        vp_overrides = self._acquire_overrides(vp_handle)
        vp_overrides.linetype = value

    def get_linetype(self, vp_handle: str) -> str:
        """Returns the linetype override or the original layer value if no
        override exist.
        """
        vp_overrides = self._get_overrides(vp_handle)
        return vp_overrides.linetype

    def get_lineweight(self, vp_handle: str) -> int:
        """Returns the lineweight override or the original layer value if no
        override exist.
        """
        vp_overrides = self._get_overrides(vp_handle)
        return vp_overrides.lineweight

    def set_lineweight(self, vp_handle: str, value: int) -> None:
        """Set the lineweight override.

        Raises:
            ValueError: invalid lineweight value
        """
        if not is_valid_layer_lineweight(value):
            raise ValueError(
                f"invalid lineweight: {value}, a linetype table entry is required"
            )
        vp_overrides = self._acquire_overrides(vp_handle)
        vp_overrides.lineweight = value

    def discard(self, vp_handle: Optional[str] = None) -> None:
        """Discard all attribute overrides for the given :class:`Viewport`
        handle or for all :class:`Viewport` entities if the handle is ``None``.
        """
        if vp_handle is None:
            self._overrides.clear()
            return
        try:
            del self._overrides[vp_handle]
        except KeyError:
            pass


def default_ovr_settings(layer) -> OverrideAttributes:
    """Returns the default settings of the layer."""
    return OverrideAttributes(
        aci=layer.color,
        rgb=layer.rgb,
        transparency=layer.transparency,
        linetype=layer.dxf.linetype,
        lineweight=layer.dxf.lineweight,
    )


def is_layer_frozen_in_vp(layer, vp_handle) -> bool:
    """Returns ``True`` if layer is frozen in VIEWPORT defined by the vp_handle."""
    vp = cast("Viewport", layer.doc.entitydb.get(vp_handle))
    if vp is not None:
        return layer.dxf.name in vp.frozen_layers
    return False


def load_layer_overrides(layer: Layer) -> dict[str, OverrideAttributes]:
    """Load all VIEWPORT overrides from the layer extension dictionary."""

    def get_ovr(vp_handle: str):
        ovr = overrides.get(vp_handle)
        if ovr is None:
            ovr = default_ovr_settings(layer)
            overrides[vp_handle] = ovr
        return ovr

    def set_alpha(vp_handle: str, value: int):
        ovr = get_ovr(vp_handle)
        ovr.transparency = clr.transparency2float(value)

    def set_color(vp_handle: str, value: int):
        ovr = get_ovr(vp_handle)
        type_, data = clr.decode_raw_color(value)
        if type_ == clr.COLOR_TYPE_ACI:
            ovr.aci = data
        elif type_ == clr.COLOR_TYPE_RGB:
            ovr.rgb = data

    def set_ltype(vp_handle: str, lt_handle: str):
        ltype = entitydb.get(lt_handle)
        if ltype is not None:
            ovr = get_ovr(vp_handle)
            ovr.linetype = ltype.dxf.name

    def set_lw(vp_handle: str, value: int):
        ovr = get_ovr(vp_handle)
        ovr.lineweight = value

    def set_xdict_state():
        xdict = layer.get_extension_dict()
        for key, code, setter in [
            (const.OVR_ALPHA_KEY, const.OVR_ALPHA_CODE, set_alpha),
            (const.OVR_COLOR_KEY, const.OVR_COLOR_CODE, set_color),
            (const.OVR_LTYPE_KEY, const.OVR_LTYPE_CODE, set_ltype),
            (const.OVR_LW_KEY, const.OVR_LW_CODE, set_lw),
        ]:
            xrec = cast("XRecord", xdict.get(key))
            if xrec is not None:
                for vp_handle, value in _load_ovr_values(xrec, code):
                    setter(vp_handle, value)

    assert layer.doc is not None, "valid DXF document required"
    entitydb: EntityDB = layer.doc.entitydb
    assert entitydb is not None, "valid entity database required"

    overrides: dict[str, OverrideAttributes] = dict()
    if not layer.has_extension_dict:
        return overrides

    set_xdict_state()
    return overrides


def _load_ovr_values(xrec: XRecord, group_code):
    tags = xrec.tags
    handles = [value for code, value in tags.find_all(const.OVR_VP_HANDLE_CODE)]
    values = [value for code, value in tags.find_all(group_code)]
    return zip(handles, values)


def store_layer_overrides(
    layer: Layer, overrides: dict[str, OverrideAttributes]
) -> None:
    """Store all VIEWPORT overrides in the layer extension dictionary.
    Replaces all existing overrides!
    """
    from ezdxf.lldxf.types import DXFTag

    def get_xdict():
        if layer.has_extension_dict:
            return layer.get_extension_dict()
        else:
            return layer.new_extension_dict()

    def set_xdict_tags(key: str, tags: list[DXFTag]):
        from ezdxf.entities import XRecord

        xdict = get_xdict()
        xrec = xdict.get(key)
        if not isinstance(xrec, XRecord) and xrec is not None:
            logger.debug(
                f"Found entity {str(xrec)} as override storage in {str(layer)} "
                f"but expected XRECORD"
            )
            xrec = None
        if xrec is None:
            xrec = xdict.add_xrecord(key)
            xrec.dxf.cloning = 1
        xrec.reset(tags)

    def del_xdict_tags(key: str):
        if not layer.has_extension_dict:
            return
        xdict = layer.get_extension_dict()
        xrec = xdict.get(key)
        if xrec is not None:
            xrec.destroy()
            xdict.discard(key)

    def make_tags(data: list[tuple[Any, str]], name: str, code: int) -> list[DXFTag]:
        tags: list[DXFTag] = []
        for value, vp_handle in data:
            tags.extend(
                [
                    DXFTag(102, name),
                    DXFTag(const.OVR_VP_HANDLE_CODE, vp_handle),
                    DXFTag(code, value),
                    DXFTag(102, "}"),
                ]
            )
        return tags

    def collect_alphas():
        for vp_handle, ovr in vp_exist.items():
            if ovr.transparency != default.transparency:
                yield clr.float2transparency(ovr.transparency), vp_handle

    def collect_colors():
        for vp_handle, ovr in vp_exist.items():
            if ovr.aci != default.aci or ovr.rgb != default.rgb:
                if ovr.rgb is None:
                    raw_color = clr.encode_raw_color(ovr.aci)
                else:
                    raw_color = clr.encode_raw_color(ovr.rgb)
                yield raw_color, vp_handle

    def collect_linetypes():
        for vp_handle, ovr in vp_exist.items():
            if ovr.linetype != default.linetype:
                ltype = layer.doc.linetypes.get(ovr.linetype)
                if ltype is not None:
                    yield ltype.dxf.handle, vp_handle

    def collect_lineweights():
        for vp_handle, ovr in vp_exist.items():
            if ovr.lineweight != default.lineweight:
                yield ovr.lineweight, vp_handle

    assert layer.doc is not None, "valid DXF document required"
    entitydb = layer.doc.entitydb
    vp_exist = {
        vp_handle: ovr
        for vp_handle, ovr in overrides.items()
        if (vp_handle in entitydb) and entitydb[vp_handle].is_alive
    }
    default = default_ovr_settings(layer)
    alphas = list(collect_alphas())
    if alphas:
        tags = make_tags(alphas, const.OVR_APP_ALPHA, const.OVR_ALPHA_CODE)
        set_xdict_tags(const.OVR_ALPHA_KEY, tags)
    else:
        del_xdict_tags(const.OVR_ALPHA_KEY)

    colors = list(collect_colors())
    if colors:
        tags = make_tags(colors, const.OVR_APP_COLOR, const.OVR_COLOR_CODE)
        set_xdict_tags(const.OVR_COLOR_KEY, tags)
    else:
        del_xdict_tags(const.OVR_COLOR_KEY)

    linetypes = list(collect_linetypes())
    if linetypes:
        tags = make_tags(linetypes, const.OVR_APP_LTYPE, const.OVR_LTYPE_CODE)
        set_xdict_tags(const.OVR_LTYPE_KEY, tags)
    else:
        del_xdict_tags(const.OVR_LTYPE_KEY)

    lineweights = list(collect_lineweights())
    if lineweights:
        tags = make_tags(lineweights, const.OVR_APP_LW, const.OVR_LW_CODE)
        set_xdict_tags(const.OVR_LW_KEY, tags)
    else:
        del_xdict_tags(const.OVR_LW_KEY)



# Copyright (c) 2011-2020, Manfred Moitzi
# License: MIT License
import pytest

from ezdxf.lldxf.const import VTX_3D_POLYLINE_VERTEX
from ezdxf import DXFIndexError
from ezdxf.layouts import VirtualLayout


@pytest.fixture
def msp():
    return VirtualLayout()


def test_create_polyline2D(msp):
    polyline = msp.add_polyline2d([(0, 0), (1, 1)])
    assert (0.0, 0.0) == polyline[0].dxf.location
    assert (1.0, 1.0) == polyline[1].dxf.location
    assert "AcDb2dPolyline" == polyline.get_mode()


def test_create_polyline3D(msp):
    polyline = msp.add_polyline3d([(1, 2, 3), (4, 5, 6)])
    assert (1.0, 2.0, 3.0) == polyline[0].dxf.location
    assert (4.0, 5.0, 6.0) == polyline[1].dxf.location
    assert VTX_3D_POLYLINE_VERTEX == polyline[0].dxf.flags
    assert "AcDb3dPolyline" == polyline.get_mode()


def test_polyline3d_vertex_layer(msp):
    attribs = {"layer": "polyline_layer"}
    polyline = msp.add_polyline3d([(1, 2, 3), (4, 5, 6)], dxfattribs=attribs)
    for vertex in polyline.vertices:
        assert (
            "polyline_layer" == vertex.dxf.layer
        ), "VERTEX entity not on the same layer as the POLYLINE entity."


def test_polyline3d_change_polyline_layer(msp):
    attribs = {"layer": "polyline_layer"}
    polyline = msp.add_polyline3d([(1, 2, 3), (4, 5, 6)], dxfattribs=attribs)
    polyline.dxf.layer = "changed_layer"
    for vertex in polyline.vertices:
        assert (
            "changed_layer" == vertex.dxf.layer
        ), "VERTEX entity not on the same layer as the POLYLINE entity."


def test_polyline2d_set_vertex(msp):
    polyline = msp.add_polyline2d([(0, 0), (1, 1), (2, 2), (3, 3)])
    polyline[2].dxf.location = (7, 7)
    assert (7.0, 7.0) == polyline[2].dxf.location


def test_polyline2d_points(msp):
    points = [(0, 0), (1, 1), (2, 2), (3, 3)]
    polyline = msp.add_polyline2d(points)
    assert points == list(polyline.points())


def test_polyline2d_point_slicing(msp):
    points = [(0, 0), (1, 1), (2, 2), (3, 3)]
    polyline = msp.add_polyline2d(points)
    assert [(1, 1), (2, 2)] == list(polyline.points())[1:3]


def test_poyline2d_append_vertices(msp):
    polyline = msp.add_polyline2d([(0, 0), (1, 1)])
    polyline.append_vertices([(7, 7), (8, 8)])
    assert (7.0, 7.0) == polyline[2].dxf.location
    assert 4 == len(polyline)


def test_polyline2d_insert_vertices(msp):
    polyline = msp.add_polyline2d([(0, 0), (1, 1)])
    polyline.insert_vertices(0, [(7, 7), (8, 8)])
    assert (7, 7) == polyline[0].dxf.location
    assert (1, 1) == polyline[3].dxf.location
    assert 4 == len(polyline)


def test_polyline2d_delete_one_vertex(msp):
    polyline = msp.add_polyline2d([(0, 0), (1, 1), (2, 2), (3, 3)])
    del polyline.vertices[0]
    assert (1, 1) == polyline[0].dxf.location
    assert 3 == len(polyline)


def test_polyline2d_delete_two_vertices(msp):
    polyline = msp.add_polyline2d([(0, 0), (1, 1), (2, 2), (3, 3)])
    del polyline.vertices[0:2]
    assert (2, 2) == polyline[0].dxf.location
    assert 2 == len(polyline)


def test_polymesh_create_mesh(msp):
    msp.add_polymesh((4, 4))
    assert True


def test_polymesh_set_vertex(msp):
    mesh = msp.add_polymesh((4, 4))
    mesh.set_mesh_vertex((1, 1), (1, 2, 3))
    result = mesh.get_mesh_vertex((1, 1)).dxf.location
    assert (1, 2, 3) == result


def test_polymesh_error_nindex(msp):
    mesh = msp.add_polymesh((4, 4))
    with pytest.raises(DXFIndexError):
        mesh.get_mesh_vertex((0, 4))


def test_polymesh_error_mindex(msp):
    mesh = msp.add_polymesh((4, 4))
    with pytest.raises(DXFIndexError):
        mesh.get_mesh_vertex((4, 0))


def test_polymesh_mesh_cache(msp):
    pos = (2, 1)
    mesh = msp.add_polymesh((4, 4))
    cache = mesh.get_mesh_vertex_cache()
    cache[pos] = (1, 2, 3)
    vertex = mesh.get_mesh_vertex(pos)
    assert vertex.dxf.location == cache[pos]
    with pytest.raises(DXFIndexError):
        cache[4, 0]


def test_polyface_create_face(msp):
    face = msp.add_polyface()
    assert 0 == len(face)


def test_polyface_add_face(msp):
    face = msp.add_polyface()
    face.append_face([(0, 0), (1, 1), (2, 2), (3, 3)])
    assert [(0, 0), (1, 1), (2, 2), (3, 3), (0, 0, 0)] == list(face.points())


def test_polyface_face_indices(msp):
    face = msp.add_polyface()
    face.append_face([(0, 0), (1, 1), (2, 2), (3, 3)])
    face_record = face[4]
    assert 1 == face_record.dxf.vtx0
    assert 2 == face_record.dxf.vtx1
    assert 3 == face_record.dxf.vtx2
    assert 4 == face_record.dxf.vtx3


def test_polyface_add_two_face_indices(msp):
    face = msp.add_polyface()
    face.append_face([(0, 0), (1, 1), (2, 2), (3, 3)])
    # second face has same vertices as the first face
    face.append_face([(0, 0), (1, 1), (2, 2)])
    face_record = face[5]  # second face
    assert 1 == face_record.dxf.vtx0
    assert 2 == face_record.dxf.vtx1
    assert 3 == face_record.dxf.vtx2
    assert 4 == face.dxf.m_count  # vertices count
    assert 2 == face.dxf.n_count  # faces count


def test_polyface_faces(msp):
    face = msp.add_polyface()
    face.append_face([(0, 0), (1, 1), (2, 2), (3, 3)])
    face.append_face([(0, 0), (1, 1), (2, 2)])
    result = list(face.faces())
    assert 2 == len(result)
    points1 = [vertex.dxf.location for vertex in result[0]]
    # the last vertex is the face_record and is always (0, 0, 0)
    # the face_record contains indices to the face building vertices
    assert [(0, 0, 0), (1, 1, 0), (2, 2, 0), (3, 3, 0), (0, 0, 0)] == points1


def test_polyface_optimized_cube(msp):
    face = msp.add_polyface()
    # a cube consist of 6 faces and 24 vertices
    # duplicated vertices should be removed
    face.append_faces(cube_faces())
    assert 8 == face.dxf.m_count  # vertices count
    assert 6 == face.dxf.n_count  # faces count


def cube_faces():
    # cube corner points
    p1 = (0, 0, 0)
    p2 = (0, 0, 1)
    p3 = (0, 1, 0)
    p4 = (0, 1, 1)
    p5 = (1, 0, 0)
    p6 = (1, 0, 1)
    p7 = (1, 1, 0)
    p8 = (1, 1, 1)

    # define the 6 cube faces
    # look into -x direction
    # Every add_face adds 4 vertices 6x4 = 24 vertices
    return [
        [p1, p5, p7, p3],
        [p1, p5, p6, p2],
        [p5, p7, p8, p6],
        [p7, p8, p4, p3],
        [p1, p3, p4, p2],
        [p2, p6, p8, p4],
    ]



# Copyright (c) 2019 Manfred Moitzi
# License: MIT License
import pytest
import ezdxf

LAYER = "test"


@pytest.fixture
def dxf():
    doc = ezdxf.new("R2000")
    doc.layers.new(LAYER)
    msp = doc.modelspace()
    msp.add_text("A", dxfattribs={"layer": LAYER})
    blk = doc.blocks.new("TEST")
    blk.add_text("A", dxfattribs={"layer": LAYER})
    return doc


def test_rename_layer(dxf):
    NEW_LAYER = "XYZ"
    layer = dxf.layers.get(LAYER)
    layer.rename(NEW_LAYER)

    assert dxf.layers.has_entry(NEW_LAYER)

    msp = dxf.modelspace()
    assert msp[0].dxf.layer == NEW_LAYER

    blk = dxf.blocks.get("TEST")
    assert blk[0].dxf.layer == NEW_LAYER


def test_rename_layer_errors(dxf):
    layer = dxf.layers.get(LAYER)
    pytest.raises(ValueError, layer.rename, "XXX*")
    pytest.raises(ValueError, layer.rename, LAYER)

    layer = dxf.layers.get("0")
    pytest.raises(ValueError, layer.rename, LAYER)

    layer = dxf.layers.get("DEFPOINTS")
    pytest.raises(ValueError, layer.rename, LAYER)



# Copyright (c) 2019-2022 Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import TYPE_CHECKING, Optional
from .dxfns import SubclassProcessor, DXFNamespace
from .dxfentity import DXFEntity
from ezdxf.lldxf.attributes import (
    DXFAttr,
    DXFAttributes,
    DefSubclass,
    group_code_mapping,
)
from ezdxf.lldxf.const import DXF2004, DXF2000
from .factory import register_entity

if TYPE_CHECKING:
    from ezdxf.document import Drawing
    from ezdxf.lldxf.tagwriter import AbstractTagWriter
    from ezdxf.lldxf.extendedtags import ExtendedTags

__all__ = ["DXFClass"]

class_def = DefSubclass(
    None,
    {
        # Class DXF record name; always unique
        "name": DXFAttr(1),
        # C++ class name. Used to bind with software that defines object class
        # behavior; always unique
        "cpp_class_name": DXFAttr(2),
        # Application name. Posted in Alert box when a class definition listed in
        # this section is not currently loaded
        "app_name": DXFAttr(3),
        # Proxy capabilities flag. Bit-coded value that indicates the capabilities
        # of this object as a proxy:
        # 0 = No operations allowed (0)
        # 1 = Erase allowed (0x1)
        # 2 = Transform allowed (0x2)
        # 4 = Color change allowed (0x4)
        # 8 = Layer change allowed (0x8)
        # 16 = Linetype change allowed (0x10)
        # 32 = Linetype scale change allowed (0x20)
        # 64 = Visibility change allowed (0x40)
        # 128 = Cloning allowed (0x80)
        # 256 = Lineweight change allowed (0x100)
        # 512 = Plot Style Name change allowed (0x200)
        # 895 = All operations except cloning allowed (0x37F)
        # 1023 = All operations allowed (0x3FF)
        # 1024 = Disables proxy warning dialog (0x400)
        # 32768 = R13 format proxy (0x8000)
        "flags": DXFAttr(90, default=0),
        # Instance count for a custom class
        "instance_count": DXFAttr(91, dxfversion=DXF2004, default=0),
        # Was-a-proxy flag. Set to 1 if class was not loaded when this DXF file was
        # created, and 0 otherwise
        "was_a_proxy": DXFAttr(280, default=0),
        # Is-an-entity flag. Set to 1 if class was derived from the AcDbEntity class
        # and can reside in the BLOCKS or ENTITIES section. If 0, instances may
        # appear only in the OBJECTS section
        "is_an_entity": DXFAttr(281, default=0),
    },
)
class_def_group_codes = group_code_mapping(class_def)


@register_entity
class DXFClass(DXFEntity):
    DXFTYPE = "CLASS"
    DXFATTRIBS = DXFAttributes(class_def)
    MIN_DXF_VERSION_FOR_EXPORT = DXF2000

    @classmethod
    def new(
        cls,
        handle: Optional[str] = None,
        owner: Optional[str] = None,
        dxfattribs=None,
        doc: Optional[Drawing] = None,
    ) -> DXFClass:
        """New CLASS constructor - has no handle, no owner and do not need
        document reference .
        """
        dxf_class = cls()
        dxf_class.doc = doc
        dxfattribs = dxfattribs or {}
        dxf_class.update_dxf_attribs(dxfattribs)
        return dxf_class

    def load_tags(
        self, tags: ExtendedTags, dxfversion: Optional[str] = None
    ) -> None:
        """Called by load constructor. CLASS is special."""
        if tags:
            # do not process base class!!!
            self.dxf = DXFNamespace(entity=self)
            processor = SubclassProcessor(tags)
            processor.fast_load_dxfattribs(
                self.dxf, class_def_group_codes, 0, log=False
            )

    def export_dxf(self, tagwriter: AbstractTagWriter):
        """Do complete export here, because CLASS is special."""
        dxfversion = tagwriter.dxfversion
        if dxfversion < DXF2000:
            return
        attribs = self.dxf
        tagwriter.write_tag2(0, self.DXFTYPE)
        attribs.export_dxf_attribs(
            tagwriter,
            [
                "name",
                "cpp_class_name",
                "app_name",
                "flags",
                "instance_count",
                "was_a_proxy",
                "is_an_entity",
            ],
        )

    @property
    def key(self) -> tuple[str, str]:
        return self.dxf.name, self.dxf.cpp_class_name



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License

import pytest
from ezdxf.tools.text import MTextEditor


def test_append_text():
    m = MTextEditor()
    m.append("TEXT")
    assert str(m) == "TEXT"


def test_iadd_text():
    m = MTextEditor()
    m += "TEXT"
    assert str(m) == "TEXT"


def test_stacked_text_limits_style():
    m = MTextEditor().stack("1", "2")
    assert str(m) == r"\S1^ 2;"


def test_stacked_text_with_horizontal_divider_line():
    m = MTextEditor().stack("1", "2", "/")
    assert str(m) == r"\S1/2;"  # no space after "/" required


def test_stacked_text_with_slanted_divider_line():
    m = MTextEditor().stack("1", "2", "#")
    assert str(m) == r"\S1#2;"  # no space after "#" required


def test_invalid_divider_char_raises_value_error():
    m = MTextEditor()
    pytest.raises(ValueError, m.stack, "1", "2", "x")


def test_change_color_name():
    m = MTextEditor()
    m.color("red")
    assert str(m) == r"\C1;"
    m.clear()
    m.aci(0)
    assert str(m) == r"\C0;"


def test_change_aci_color():
    m = MTextEditor()
    m.aci(0).aci(256)
    assert str(m) == r"\C0;\C256;"


@pytest.mark.parametrize("aci", [-1, 257])
def test_aci_color_raises_value_error(aci):
    with pytest.raises(ValueError):
        MTextEditor().aci(aci)


def test_change_to_red_by_rgb():
    m = MTextEditor().rgb((255, 0, 0))
    assert str(m) == r"\c255;"


def test_change_to_green_by_rgb():
    m = MTextEditor().rgb((0, 255, 0))
    assert str(m) == r"\c65280;"


def test_change_to_blue_by_rgb():
    m = MTextEditor().rgb((0, 0, 255))
    assert str(m) == r"\c16711680;"


def test_change_font():
    m = MTextEditor()
    m.font("Arial", bold=False, italic=False)
    assert str(m) == r"\fArial|b0|i0;"


def test_scale_height_factor():
    assert str(MTextEditor().scale_height(2)) == r"\H2x;"
    assert str(MTextEditor().scale_height(1.6666)) == r"\H1.667x;"


def test_absolute_text_height():
    assert str(MTextEditor().height(2)) == r"\H2;"
    assert str(MTextEditor().height(1.6666)) == r"\H1.667;"


def test_change_width_factor():
    assert str(MTextEditor().width_factor(2)) == r"\W2;"
    assert str(MTextEditor().width_factor(1.6666)) == r"\W1.667;"


def test_change_char_tracking_factor():
    assert str(MTextEditor().char_tracking_factor(2)) == r"\T2;"
    assert str(MTextEditor().char_tracking_factor(1.6666)) == r"\T1.667;"


def test_change_oblique_angle():
    assert str(MTextEditor().oblique(0)) == r"\Q0;"  # vertical
    assert str(MTextEditor().oblique(15)) == r"\Q15;"


def test_fluent_interface():
    m = MTextEditor("some text").color("red").stack("1", "2").append("end.")
    assert str(m) == r"some text\C1;\S1^ 2;end."


def test_grouping():
    m = MTextEditor("some text")
    group_content = str(MTextEditor().font("Arial").append("Font=Arial"))
    m.group(group_content)
    assert str(m) == r"some text{\fArial|b0|i0;Font=Arial}"


def test_underline_text():
    assert str(MTextEditor().underline("TEXT")) == r"\LTEXT\l"


def test_overline_text():
    assert str(MTextEditor().overline("TEXT")) == r"\OTEXT\o"


def test_strike_through_text():
    assert str(MTextEditor().strike_through("TEXT")) == r"\KTEXT\k"


def test_bullet_lists():
    result = MTextEditor().bullet_list(
        indent=4,  # left indentation of the list items
        bullets=["-", "+"],  # bullets - mark in front of the list item
        content=["first", "second"],  # list items
    )
    assert str(result) == r"{\pxi-3,l4,t4;-^Ifirst\P+^Isecond\P}"


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2018-2024, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import (
    Tuple,
    Iterable,
    Sequence,
    TYPE_CHECKING,
    Iterator,
    Optional,
)
from functools import partial
import math
import random

if TYPE_CHECKING:
    from ezdxf.math import UVec, AnyVec

ABS_TOL = 1e-12
isclose = partial(math.isclose, abs_tol=ABS_TOL)

__all__ = ["Vec3", "Vec2"]


class Vec3:
    """Immutable 3D vector class.

    This class is optimized for universality not for speed.
    Immutable means you can't change (x, y, z) components after initialization::

        v1 = Vec3(1, 2, 3)
        v2 = v1
        v2.z = 7  # this is not possible, raises AttributeError
        v2 = Vec3(v2.x, v2.y, 7)  # this creates a new Vec3() object
        assert v1.z == 3  # and v1 remains unchanged


    :class:`Vec3` initialization:

        - ``Vec3()``, returns ``Vec3(0, 0, 0)``
        - ``Vec3((x, y))``, returns ``Vec3(x, y, 0)``
        - ``Vec3((x, y, z))``, returns ``Vec3(x, y, z)``
        - ``Vec3(x, y)``, returns ``Vec3(x, y, 0)``
        - ``Vec3(x, y, z)``, returns  ``Vec3(x, y, z)``

    Addition, subtraction, scalar multiplication and scalar division left and
    right-handed are supported::

        v = Vec3(1, 2, 3)
        v + (1, 2, 3) == Vec3(2, 4, 6)
        (1, 2, 3) + v == Vec3(2, 4, 6)
        v - (1, 2, 3) == Vec3(0, 0, 0)
        (1, 2, 3) - v == Vec3(0, 0, 0)
        v * 3 == Vec3(3, 6, 9)
        3 * v == Vec3(3, 6, 9)
        Vec3(3, 6, 9) / 3 == Vec3(1, 2, 3)
        -Vec3(1, 2, 3) == (-1, -2, -3)

    Comparison between vectors and vectors or tuples is supported::

        Vec3(1, 2, 3) < Vec3 (2, 2, 2)
        (1, 2, 3) < tuple(Vec3(2, 2, 2))  # conversion necessary
        Vec3(1, 2, 3) == (1, 2, 3)

        bool(Vec3(1, 2, 3)) is True
        bool(Vec3(0, 0, 0)) is False

    """

    __slots__ = ["_x", "_y", "_z"]

    def __init__(self, *args):
        self._x, self._y, self._z = self.decompose(*args)

    @property
    def x(self) -> float:
        """x-axis value"""
        return self._x

    @property
    def y(self) -> float:
        """y-axis value"""
        return self._y

    @property
    def z(self) -> float:
        """z-axis value"""
        return self._z

    @property
    def xy(self) -> Vec3:
        """Vec3 as ``(x, y, 0)``, projected on the xy-plane."""
        return self.__class__(self._x, self._y)

    @property
    def xyz(self) -> tuple[float, float, float]:
        """Vec3 as ``(x, y, z)`` tuple."""
        return self._x, self._y, self._z

    @property
    def vec2(self) -> Vec2:
        """Real 2D vector as :class:`Vec2` object."""
        return Vec2((self._x, self._y))

    def replace(
        self,
        x: Optional[float] = None,
        y: Optional[float] = None,
        z: Optional[float] = None,
    ) -> Vec3:
        """Returns a copy of vector with replaced x-, y- and/or z-axis."""
        if x is None:
            x = self._x
        if y is None:
            y = self._y
        if z is None:
            z = self._z
        return self.__class__(x, y, z)

    def round(self, ndigits=None) -> Vec3:
        """Returns a new vector where all components are rounded to `ndigits`.

        Uses standard Python :func:`round` function for rounding.
        """
        return self.__class__(
            round(self._x, ndigits),
            round(self._y, ndigits),
            round(self._z, ndigits),
        )

    @classmethod
    def list(cls, items: Iterable[UVec]) -> list[Vec3]:
        """Returns a list of :class:`Vec3` objects."""
        return list(cls.generate(items))

    @classmethod
    def tuple(cls, items: Iterable[UVec]) -> Sequence[Vec3]:
        """Returns a tuple of :class:`Vec3` objects."""
        return tuple(cls.generate(items))

    @classmethod
    def generate(cls, items: Iterable[UVec]) -> Iterator[Vec3]:
        """Returns an iterable of :class:`Vec3` objects."""
        return (cls(item) for item in items)

    @classmethod
    def from_angle(cls, angle: float, length: float = 1.0) -> Vec3:
        """Returns a :class:`Vec3` object from `angle` in radians in the
        xy-plane, z-axis = ``0``.
        """
        return cls(math.cos(angle) * length, math.sin(angle) * length, 0.0)

    @classmethod
    def from_deg_angle(cls, angle: float, length: float = 1.0) -> Vec3:
        """Returns a :class:`Vec3` object from `angle` in degrees in the
        xy-plane, z-axis = ``0``.
        """
        return cls.from_angle(math.radians(angle), length)

    @staticmethod
    def decompose(*args) -> Tuple[float, float, float]:  # cannot use "tuple" here!
        """Converts input into a (x, y, z) tuple.

        Valid arguments are:

            - no args: ``decompose()`` returns (0, 0, 0)
            - 1 arg: ``decompose(arg)``, `arg` is tuple or list, tuple has to be
              (x, y[, z]): ``decompose((x, y))`` returns (x, y, 0.)
            - 2 args: ``decompose(x, y)`` returns (x, y, 0)
            - 3 args: ``decompose(x, y, z)`` returns (x, y, z)

        Returns:
            (x, y, z) tuple

        (internal API)

        """
        length = len(args)
        if length == 0:
            return 0.0, 0.0, 0.0
        elif length == 1:
            data = args[0]
            if isinstance(data, Vec3):
                return data._x, data._y, data._z
            else:
                length = len(data)
                if length == 2:
                    x, y = data
                    z = 0.0
                elif length == 3:
                    x, y, z = data
                else:
                    raise TypeError
                return float(x), float(y), float(z)
        elif length == 2:
            x, y = args
            return float(x), float(y), 0.0
        elif length == 3:
            x, y, z = args
            return float(x), float(y), float(z)
        raise TypeError

    @classmethod
    def random(cls, length: float = 1) -> Vec3:
        """Returns a random vector."""
        x = random.uniform(-1, 1)
        y = random.uniform(-1, 1)
        z = random.uniform(-1, 1)
        return Vec3(x, y, z).normalize(length)

    def __str__(self) -> str:
        """Return ``'(x, y, z)'`` as string."""
        return "({0.x}, {0.y}, {0.z})".format(self)

    def __repr__(self) -> str:
        """Return ``'Vec3(x, y, z)'`` as string."""
        return "Vec3" + self.__str__()

    def __len__(self) -> int:
        """Returns always ``3``."""
        return 3

    def __hash__(self) -> int:
        """Returns hash value of vector, enables the usage of vector as key in
        ``set`` and ``dict``.
        """
        return hash(self.xyz)

    def copy(self) -> Vec3:
        """Returns a copy of vector as :class:`Vec3` object."""
        return self  # immutable!

    __copy__ = copy

    def __deepcopy__(self, memodict: dict) -> Vec3:
        """:func:`copy.deepcopy` support."""
        return self  # immutable!

    def __getitem__(self, index: int) -> float:
        """Support for indexing:

        - v[0] is v.x
        - v[1] is v.y
        - v[2] is v.z

        """
        if isinstance(index, slice):
            raise TypeError("slicing not supported")
        if index == 0:
            return self._x
        elif index == 1:
            return self._y
        elif index == 2:
            return self._z
        else:
            raise IndexError(f"invalid index {index}")

    def __iter__(self) -> Iterator[float]:
        """Returns iterable of x-, y- and z-axis."""
        yield self._x
        yield self._y
        yield self._z

    def __abs__(self) -> float:
        """Returns length (magnitude) of vector."""
        return self.magnitude

    @property
    def magnitude(self) -> float:
        """Length of vector."""
        return self.magnitude_square**0.5

    @property
    def magnitude_xy(self) -> float:
        """Length of vector in the xy-plane."""
        return math.hypot(self._x, self._y)

    @property
    def magnitude_square(self) -> float:
        """Square length of vector."""
        x, y, z = self._x, self._y, self._z
        return x * x + y * y + z * z

    @property
    def is_null(self) -> bool:
        """``True`` if all components are close to zero: ``Vec3(0, 0, 0)``.
        Has a fixed absolute testing tolerance of 1e-12!
        """
        return (
            abs(self._x) <= ABS_TOL
            and abs(self._y) <= ABS_TOL
            and abs(self._z) <= ABS_TOL
        )

    def is_parallel(
        self, other: Vec3, *, rel_tol: float = 1e-9, abs_tol: float = 1e-12
    ) -> bool:
        """Returns ``True`` if `self` and `other` are parallel to vectors."""
        v1 = self.normalize()
        v2 = other.normalize()
        return v1.isclose(v2, rel_tol=rel_tol, abs_tol=abs_tol) or v1.isclose(
            -v2, rel_tol=rel_tol, abs_tol=abs_tol
        )

    @property
    def spatial_angle(self) -> float:
        """Spatial angle between vector and x-axis in radians."""
        return math.acos(X_AXIS.dot(self.normalize()))

    @property
    def spatial_angle_deg(self) -> float:
        """Spatial angle between vector and x-axis in degrees."""
        return math.degrees(self.spatial_angle)

    @property
    def angle(self) -> float:
        """Angle between vector and x-axis in the xy-plane in radians."""
        return math.atan2(self._y, self._x)

    @property
    def angle_deg(self) -> float:
        """Returns angle of vector and x-axis in the xy-plane in degrees."""
        return math.degrees(self.angle)

    def orthogonal(self, ccw: bool = True) -> Vec3:
        """Returns orthogonal 2D vector, z-axis is unchanged.

        Args:
            ccw: counter-clockwise if ``True`` else clockwise

        """
        return (
            self.__class__(-self._y, self._x, self._z)
            if ccw
            else self.__class__(self._y, -self._x, self._z)
        )

    def lerp(self, other: UVec, factor=0.5) -> Vec3:
        """Returns linear interpolation between `self` and `other`.

        Args:
            other: end point as :class:`Vec3` compatible object
            factor: interpolation factor (0 = self, 1 = other,
                0.5 = mid point)

        """
        d = (self.__class__(other) - self) * float(factor)
        return self.__add__(d)

    def project(self, other: UVec) -> Vec3:
        """Returns projected vector of `other` onto `self`."""
        uv = self.normalize()
        return uv * uv.dot(other)

    def normalize(self, length: float = 1.0) -> Vec3:
        """Returns normalized vector, optional scaled by `length`."""
        return self.__mul__(length / self.magnitude)

    def reversed(self) -> Vec3:
        """Returns negated vector (-`self`)."""
        return self.__class__(-self._x, -self._y, -self._z)

    __neg__ = reversed

    def __bool__(self) -> bool:
        """Returns ``True`` if vector is not (0, 0, 0)."""
        return not self.is_null

    def isclose(
        self, other: UVec, *, rel_tol: float = 1e-9, abs_tol: float = 1e-12
    ) -> bool:
        """Returns ``True`` if `self` is close to `other`.
        Uses :func:`math.isclose` to compare all axis.

        Learn more about the :func:`math.isclose` function in
        `PEP 485 <https://www.python.org/dev/peps/pep-0485/>`_.

        """
        x, y, z = self.decompose(other)
        return (
            math.isclose(self._x, x, rel_tol=rel_tol, abs_tol=abs_tol)
            and math.isclose(self._y, y, rel_tol=rel_tol, abs_tol=abs_tol)
            and math.isclose(self._z, z, rel_tol=rel_tol, abs_tol=abs_tol)
        )

    def __eq__(self, other: UVec) -> bool:
        """Equal operator.

        Args:
            other: :class:`Vec3` compatible object
        """
        if not isinstance(other, Vec3):
            other = Vec3(other)
        return self.x == other.x and self.y == other.y and self.z == other.z

    def __lt__(self, other: UVec) -> bool:
        """Lower than operator.

        Args:
            other: :class:`Vec3` compatible object

        """
        x, y, z = self.decompose(other)
        if self._x == x:
            if self._y == y:
                return self._z < z
            else:
                return self._y < y
        else:
            return self._x < x

    def __add__(self, other: UVec) -> Vec3:
        """Add :class:`Vec3` operator: `self` + `other`."""
        x, y, z = self.decompose(other)
        return self.__class__(self._x + x, self._y + y, self._z + z)

    def __radd__(self, other: UVec) -> Vec3:
        """RAdd :class:`Vec3` operator: `other` + `self`."""
        return self.__add__(other)

    def __sub__(self, other: UVec) -> Vec3:
        """Sub :class:`Vec3` operator: `self` - `other`."""

        x, y, z = self.decompose(other)
        return self.__class__(self._x - x, self._y - y, self._z - z)

    def __rsub__(self, other: UVec) -> Vec3:
        """RSub :class:`Vec3` operator: `other` - `self`."""
        x, y, z = self.decompose(other)
        return self.__class__(x - self._x, y - self._y, z - self._z)

    def __mul__(self, other: float) -> Vec3:
        """Scalar Mul operator: `self` * `other`."""
        scalar = float(other)
        return self.__class__(self._x * scalar, self._y * scalar, self._z * scalar)

    def __rmul__(self, other: float) -> Vec3:
        """Scalar RMul operator: `other` * `self`."""
        return self.__mul__(other)

    def __truediv__(self, other: float) -> Vec3:
        """Scalar Div operator: `self` / `other`."""
        scalar = float(other)
        return self.__class__(self._x / scalar, self._y / scalar, self._z / scalar)

    @staticmethod
    def sum(items: Iterable[UVec]) -> Vec3:
        """Add all vectors in `items`."""
        s = NULLVEC
        for v in items:
            s += v
        return s

    def dot(self, other: UVec) -> float:
        """Dot operator: `self` . `other`

        Args:
            other: :class:`Vec3` compatible object
        """
        x, y, z = self.decompose(other)
        return self._x * x + self._y * y + self._z * z

    def cross(self, other: UVec) -> Vec3:
        """Cross operator: `self` x `other`

        Args:
            other: :class:`Vec3` compatible object
        """
        x, y, z = self.decompose(other)
        return self.__class__(
            self._y * z - self._z * y,
            self._z * x - self._x * z,
            self._x * y - self._y * x,
        )

    def distance(self, other: UVec) -> float:
        """Returns distance between `self` and `other` vector."""
        v = self.__class__(other)
        return v.__sub__(self).magnitude

    def angle_between(self, other: UVec) -> float:
        """Returns angle between `self` and `other` in radians. +angle is
        counter clockwise orientation.

        Args:
            other: :class:`Vec3` compatible object

        """
        cos_theta = self.normalize().dot(self.__class__(other).normalize())
        # avoid domain errors caused by floating point imprecision:
        if cos_theta < -1.0:
            cos_theta = -1.0
        elif cos_theta > 1.0:
            cos_theta = 1.0
        return math.acos(cos_theta)

    def angle_about(self, base: UVec, target: UVec) -> float:
        # (c) 2020 by Matt Broadway, MIT License
        """Returns counter-clockwise angle in radians about `self` from `base` to
        `target` when projected onto the plane defined by `self` as the normal
        vector.

        Args:
            base: base vector, defines angle 0
            target: target vector
        """
        x_axis = (base - self.project(base)).normalize()
        y_axis = self.cross(x_axis).normalize()
        target_projected_x = x_axis.dot(target)
        target_projected_y = y_axis.dot(target)
        return math.atan2(target_projected_y, target_projected_x) % math.tau

    def rotate(self, angle: float) -> Vec3:
        """Returns vector rotated about `angle` around the z-axis.

        Args:
            angle: angle in radians

        """
        v = self.__class__(self.x, self.y, 0.0)
        v = Vec3.from_angle(v.angle + angle, v.magnitude)
        return self.__class__(v.x, v.y, self.z)

    def rotate_deg(self, angle: float) -> Vec3:
        """Returns vector rotated about `angle` around the z-axis.

        Args:
            angle: angle in degrees

        """
        return self.rotate(math.radians(angle))


X_AXIS = Vec3(1, 0, 0)
Y_AXIS = Vec3(0, 1, 0)
Z_AXIS = Vec3(0, 0, 1)
NULLVEC = Vec3(0, 0, 0)


def distance(p1: UVec, p2: UVec) -> float:
    """Returns distance between points `p1` and `p2`.

    Args:
        p1: first point as :class:`Vec3` compatible object
        p2: second point as :class:`Vec3` compatible object

    """
    return Vec3(p1).distance(p2)


def lerp(p1: UVec, p2: UVec, factor: float = 0.5) -> Vec3:
    """Returns linear interpolation between points `p1` and `p2` as :class:`Vec3`.

    Args:
        p1: first point as :class:`Vec3` compatible object
        p2: second point as :class:`Vec3` compatible object
        factor:  interpolation factor (``0`` = `p1`, ``1`` = `p2`, ``0.5`` = mid point)

    """
    return Vec3(p1).lerp(p2, factor)


class Vec2:
    """Immutable 2D vector class.

    Args:
        v: vector object with :attr:`x` and :attr:`y` attributes/properties or a
           sequence of float ``[x, y, ...]`` or x-axis as float if argument `y`
           is not ``None``
        y: second float for :code:`Vec2(x, y)`

    :class:`Vec2` implements a subset of :class:`Vec3`.

    """

    __slots__ = ["x", "y"]

    def __init__(self, v=(0.0, 0.0), y=None) -> None:
        try:  # fast path for Vec2() and Vec3() or any object providing x and y attributes
            self.x = v.x
            self.y = v.y
        except AttributeError:
            if y is None:  # given one tuple
                self.x = float(v[0])
                self.y = float(v[1])
            else:  # two floats given
                self.x = float(v)
                self.y = float(y)

    @property
    def vec3(self) -> Vec3:
        """Returns a 3D vector."""
        return Vec3(self.x, self.y, 0)

    def round(self, ndigits=None) -> Vec2:
        """Returns a new vector where all components are rounded to `ndigits`.

        Uses standard Python :func:`round` function for rounding.
        """
        return self.__class__(round(self.x, ndigits), round(self.y, ndigits))

    @classmethod
    def list(cls, items: Iterable[UVec]) -> list[Vec2]:
        return list(cls.generate(items))

    @classmethod
    def tuple(cls, items: Iterable[UVec]) -> Sequence[Vec2]:
        """Returns a tuple of :class:`Vec3` objects."""
        return tuple(cls.generate(items))

    @classmethod
    def generate(cls, items: Iterable[UVec]) -> Iterator[Vec2]:
        return (cls(item) for item in items)

    @classmethod
    def from_angle(cls, angle: float, length: float = 1.0) -> Vec2:
        return cls(math.cos(angle) * length, math.sin(angle) * length)

    @classmethod
    def from_deg_angle(cls, angle: float, length: float = 1.0) -> Vec2:
        return cls.from_angle(math.radians(angle), length)

    def __str__(self) -> str:
        return "({0.x}, {0.y})".format(self)

    def __repr__(self) -> str:
        return "Vec2" + self.__str__()

    def __len__(self) -> int:
        return 2

    def __hash__(self) -> int:
        return hash((self.x, self.y))

    def copy(self) -> Vec2:
        return self.__class__((self.x, self.y))

    __copy__ = copy

    def __deepcopy__(self, memodict: dict) -> Vec2:
        try:
            return memodict[id(self)]
        except KeyError:
            v = self.copy()
            memodict[id(self)] = v
            return v

    def __getitem__(self, index: int) -> float:
        if isinstance(index, slice):
            raise TypeError("slicing not supported")
        if index == 0:
            return self.x
        elif index == 1:
            return self.y
        else:
            raise IndexError(f"invalid index {index}")

    def __iter__(self) -> Iterator[float]:
        yield self.x
        yield self.y

    def __abs__(self) -> float:
        return self.magnitude

    @property
    def magnitude(self) -> float:
        """Returns length of vector."""
        return math.hypot(self.x, self.y)

    @property
    def is_null(self) -> bool:
        return abs(self.x) <= ABS_TOL and abs(self.y) <= ABS_TOL

    @property
    def angle(self) -> float:
        """Angle of vector in radians."""
        return math.atan2(self.y, self.x)

    @property
    def angle_deg(self) -> float:
        """Angle of vector in degrees."""
        return math.degrees(self.angle)

    def orthogonal(self, ccw: bool = True) -> Vec2:
        """Orthogonal vector

        Args:
            ccw: counter-clockwise if ``True`` else clockwise

        """
        if ccw:
            return self.__class__(-self.y, self.x)
        else:
            return self.__class__(self.y, -self.x)

    def lerp(self, other: AnyVec, factor: float = 0.5) -> Vec2:
        """Linear interpolation between `self` and `other`.

        Args:
            other: target vector/point
            factor: interpolation factor (0=self, 1=other, 0.5=mid point)

        Returns: interpolated vector

        """
        x = self.x + (other.x - self.x) * factor
        y = self.y + (other.y - self.y) * factor
        return self.__class__(x, y)

    def project(self, other: AnyVec) -> Vec2:
        """Project vector `other` onto `self`."""
        uv = self.normalize()
        return uv * uv.dot(other)

    def normalize(self, length: float = 1.0) -> Vec2:
        return self.__mul__(length / self.magnitude)

    def reversed(self) -> Vec2:
        return self.__class__(-self.x, -self.y)

    __neg__ = reversed

    def __bool__(self) -> bool:
        return not self.is_null

    def isclose(
        self, other: AnyVec, *, rel_tol: float = 1e-9, abs_tol: float = 1e-12
    ) -> bool:
        if not isinstance(other, Vec2):
            other = Vec2(other)
        return math.isclose(
            self.x, other.x, rel_tol=rel_tol, abs_tol=abs_tol
        ) and math.isclose(self.y, other.y, rel_tol=rel_tol, abs_tol=abs_tol)

    def __eq__(self, other: UVec) -> bool:
        if not isinstance(other, Vec2):
            other = Vec2(other)
        return self.x == other.x and self.y == other.y

    def __lt__(self, other: UVec) -> bool:
        # accepts also tuples, for more convenience at testing
        x, y, *_ = other
        if self.x == x:
            return self.y < y
        else:
            return self.x < x

    def __add__(self, other: AnyVec) -> Vec2:
        try:
            return self.__class__(self.x + other.x, self.y + other.y)
        except AttributeError:
            raise TypeError("invalid argument")

    def __sub__(self, other: AnyVec) -> Vec2:
        try:
            return self.__class__(self.x - other.x, self.y - other.y)
        except AttributeError:
            raise TypeError("invalid argument")

    def __rsub__(self, other: AnyVec) -> Vec2:
        try:
            return self.__class__(other.x - self.x, other.y - self.y)
        except AttributeError:
            raise TypeError("invalid argument")

    def __mul__(self, other: float) -> Vec2:
        return self.__class__(self.x * other, self.y * other)

    def __rmul__(self, other: float) -> Vec2:
        return self.__class__(self.x * other, self.y * other)

    def __truediv__(self, other: float) -> Vec2:
        return self.__class__(self.x / other, self.y / other)

    def dot(self, other: AnyVec) -> float:
        return self.x * other.x + self.y * other.y

    def det(self, other: AnyVec) -> float:
        return self.x * other.y - self.y * other.x

    def distance(self, other: AnyVec) -> float:
        return math.hypot(self.x - other.x, self.y - other.y)

    def angle_between(self, other: AnyVec) -> float:
        """Calculate angle between `self` and `other` in radians. +angle is
        counter-clockwise orientation.

        """
        cos_theta = self.normalize().dot(other.normalize())
        # avoid domain errors caused by floating point imprecision:
        if cos_theta < -1.0:
            cos_theta = -1.0
        elif cos_theta > 1.0:
            cos_theta = 1.0
        return math.acos(cos_theta)

    def rotate(self, angle: float) -> Vec2:
        """Rotate vector around origin.

        Args:
            angle: angle in radians

        """
        return self.__class__.from_angle(self.angle + angle, self.magnitude)

    def rotate_deg(self, angle: float) -> Vec2:
        """Rotate vector around origin.

        Args:
            angle: angle in degrees

        Returns: rotated vector

        """
        return self.__class__.from_angle(
            self.angle + math.radians(angle), self.magnitude
        )

    @staticmethod
    def sum(items: Iterable[Vec2]) -> Vec2:
        """Add all vectors in `items`."""
        s = Vec2(0, 0)
        for v in items:
            s += v
        return s



#  Copyright (c) 2021-2023, Manfred Moitzi
#  License: MIT License

import pytest

pytest.importorskip("PySide6")

from ezdxf import path
from ezdxf.math import quadratic_to_cubic_bezier, Bezier3P, Vec2
from ezdxf import npshapes


class TestNumpyPath2dToQPainterPath:
    def test_no_paths(self):
        with pytest.raises(ValueError):
            npshapes.to_qpainter_path([])

    def test_line_to(self):
        p = path.Path()
        p.line_to((4, 5, 6))
        p.line_to((7, 8, 6))
        qpath = npshapes.to_qpainter_path([npshapes.NumpyPath2d(p)])
        assert qpath.elementCount() == 3

        m = qpath.elementAt(0)
        assert m.isMoveTo() is True
        assert (m.x, m.y) == (0, 0)

        l1 = qpath.elementAt(1)
        assert l1.isLineTo() is True
        assert (l1.x, l1.y) == (4, 5)

        l2 = qpath.elementAt(2)
        assert l2.isLineTo() is True
        assert (l2.x, l2.y) == (7, 8)

    def test_curve3_to(self):
        bez3 = Bezier3P(Vec2.list([(0, 0), (2, 1), (4, 0)]))
        p = path.Path()
        p.curve3_to(bez3.control_points[2], bez3.control_points[1])
        qpath = npshapes.to_qpainter_path([npshapes.NumpyPath2d(p)])
        # Qt converts quadratic bezier curves unto cubic bezier curves
        assert qpath.elementCount() == 4
        bez4 = quadratic_to_cubic_bezier(bez3)

        q1 = qpath.elementAt(1)
        assert q1.isCurveTo()  # start of cure
        assert q1.x, q1.y == bez4.control_points[1]

        q2 = qpath.elementAt(2)
        assert q2.x, q2.y == bez4.control_points[2]

        q3 = qpath.elementAt(3)
        assert q3.x, q3.y == bez4.control_points[2]

    def test_curve4_to(self):
        bez4 = [(4, 0, 2), (1, 1, 7), (3, 1, 5)]
        p = path.Path()
        p.curve4_to(*bez4)
        qpath = npshapes.to_qpainter_path([npshapes.NumpyPath2d(p)])
        assert qpath.elementCount() == 4
        q1 = qpath.elementAt(1)
        assert q1.isCurveTo()  # start of cure
        assert q1.x, q1.y == bez4[1]

        q2 = qpath.elementAt(2)
        assert q2.x, q2.y == bez4[2]

        q3 = qpath.elementAt(3)
        assert q3.x, q3.y == bez4[0]

    def test_two_single_paths(self):
        p1 = path.Path()
        p1.line_to((4, 5, 6))
        p2 = path.Path()
        p2.line_to((7, 8, 6))
        qpath = npshapes.to_qpainter_path(
            [npshapes.NumpyPath2d(p1), npshapes.NumpyPath2d(p2)]
        )
        assert qpath.elementCount() == 4
        assert qpath.elementAt(0).isMoveTo() is True
        assert qpath.elementAt(1).isLineTo() is True
        assert qpath.elementAt(2).isMoveTo() is True
        assert qpath.elementAt(3).isLineTo() is True

    def test_one_multi_paths(self):
        p = path.Path()
        p.line_to((4, 5, 6))
        p.move_to((0, 0, 0))
        p.line_to((7, 8, 6))
        qpath = npshapes.to_qpainter_path([npshapes.NumpyPath2d(p)])
        assert qpath.elementCount() == 4
        assert qpath.elementAt(0).isMoveTo() is True
        assert qpath.elementAt(1).isLineTo() is True
        assert qpath.elementAt(2).isMoveTo() is True
        assert qpath.elementAt(3).isLineTo() is True


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License

import pytest
import math

import ezdxf
from ezdxf import transform
from ezdxf.layouts import VirtualLayout


class TestInplaceMethod:
    """The tests do not check if all possible transformations will be performed correct,
    this is done by the tests for the transform() method of the entities.
    """

    @pytest.fixture
    def msp(self, request):
        if request.param == "doc":
            doc = ezdxf.new()
            msp = doc.modelspace()
        elif request.param == "virtual":
            # The matrix() function supports virtual entities as well.
            msp = VirtualLayout()
        else:
            return ValueError(request.param)
        msp.add_point((0, 0))
        msp.add_text("TEXT")
        msp.add_circle((0, 0), radius=1)
        return msp

    @pytest.mark.parametrize("msp", ["doc", "virtual"], indirect=True)
    def test_transformation_by_matrix_without_errors(self, msp):
        m = transform.Matrix44.translate(1, 2, 0)
        log = transform.inplace(msp, m)
        assert len(log) == 0
        point = msp[0]
        assert point.dxf.location.isclose((1, 2, 0))
        text = msp[1]
        assert text.dxf.insert.isclose((1, 2, 0))
        circle = msp[2]
        assert circle.dxf.center.isclose((1, 2, 0))

    @pytest.mark.parametrize("msp", ["doc"], indirect=True)
    def test_non_uniform_transformation(self, msp):
        # not supported by virtual layouts
        m = transform.Matrix44.scale(1, 2, 1)
        log = transform.inplace(msp, m)
        assert len(log) == 0
        entity = msp[2]
        assert entity.dxftype() == "ELLIPSE"

    def test_entities_without_transformation_support(self):
        from ezdxf.entities import Layer

        m = transform.Matrix44.translate(1, 0, 0)
        log = transform.inplace([Layer.new()], m)
        assert log[0].error == transform.Error.TRANSFORMATION_NOT_SUPPORTED

    def test_acis_entities(self):
        # new in v1.3.0: ACIS entities support a temporary transformation
        #
        # This way a temporary transformation of ACIS entities is stored by ezdxf.
        # This temp. transformation has to be applied before export otherwise a warning 
        # will be logged.

        msp = VirtualLayout()
        body = msp.add_body()  # ACIS entity
        m = transform.Matrix44.translate(1, 2, 3)
        log = transform.inplace(msp, m)
        assert len(log) == 0

        m2 = body.temporary_transformation().get_matrix()
        v = (3, 2, 1)
        assert m.transform(v).isclose(m2.transform(v))


class TestConvenientFunctions:
    @pytest.fixture
    def msp(self):
        msp = VirtualLayout()
        msp.add_point((1, 1, 1))
        return msp

    def test_translate(self, msp):
        transform.translate(msp, offset=(1, 2))
        point = msp[0]
        assert point.dxf.location.isclose((2, 3, 1))

    def test_scale_uniform(self, msp):
        transform.scale_uniform(msp, factor=2)
        point = msp[0]
        assert point.dxf.location.isclose((2, 2, 2))

    def test_scale(self, msp):
        transform.scale(msp, sx=2, sy=3, sz=4)
        point = msp[0]
        assert point.dxf.location.isclose((2, 3, 4))

    def test_z_rotate(self, msp):
        transform.z_rotate(msp, angle=math.pi / 2)
        point = msp[0]
        assert point.dxf.location.isclose((-1, 1, 1))

    def test_y_rotate(self, msp):
        transform.y_rotate(msp, angle=math.pi / 2)
        point = msp[0]
        assert point.dxf.location.isclose((1, 1, -1))

    def test_x_rotate(self, msp):
        transform.x_rotate(msp, angle=math.pi / 2)
        point = msp[0]
        assert point.dxf.location.isclose((1, -1, 1))


def test_circle_non_uniform_scaling():
    doc = ezdxf.new()
    msp = doc.modelspace()
    msp.add_circle((0, 0), radius=1)
    transform.scale(msp, sx=3, sy=2, sz=1)

    ellipse = msp[0]
    assert ellipse.dxftype() == "ELLIPSE"
    assert ellipse.dxf.center.isclose((0, 0, 0))
    assert ellipse.dxf.major_axis.isclose((3, 0, 0))
    assert ellipse.dxf.ratio == pytest.approx(0.6666666)


def test_polyline_non_uniform_scaling():
    doc = ezdxf.new()
    msp = doc.modelspace()
    # semi-circle: center=(0, 0); radius=1
    msp.add_lwpolyline([(-1, 0, 0, 0, 1), (1, 0)])
    transform.scale(msp, sx=3, sy=2, sz=1)

    ellipse = msp[0]
    assert ellipse.dxftype() == "ELLIPSE"
    assert ellipse.dxf.center.isclose((0, 0, 0))
    assert ellipse.dxf.major_axis.isclose((3, 0, 0))
    assert ellipse.dxf.ratio == pytest.approx(0.6666666)


def test_virtual_entities_do_not_support_non_uniform_scaling():
    msp = VirtualLayout()
    msp.add_circle((0, 0), radius=1)
    log = transform.scale(msp, sx=3, sy=2, sz=1)

    # does not transform the CIRCLE
    circle = msp[0]
    assert circle.dxftype() == "CIRCLE"
    assert circle.dxf.radius == pytest.approx(1.0)

    assert len(log) == 1
    assert log[0].error == transform.Error.VIRTUAL_ENTITY_NOT_SUPPORTED


class TestVirtualCopies:
    @pytest.fixture
    def msp(self):
        msp = VirtualLayout()
        msp.add_circle((0, 0), radius=1)
        msp.add_point((1, 1, 1))
        return msp

    def test_just_copy(self, msp):
        log, entities = transform.copies(msp)
        assert len(entities) == 2
        assert all(e.is_virtual for e in entities)

    def test_scale_virtual_circular_arcs_non_uniform(self, msp):
        log, entities = transform.copies(msp, m=transform.Matrix44.scale(2, 3, 1))
        assert len(log) == 0
        assert entities[0].dxftype() == "ELLIPSE"
        assert entities[1].dxftype() == "POINT"

    def test_scale_virtual_polyline_with_bulge_non_uniform(self):
        layout = VirtualLayout()
        layout.add_lwpolyline([(-1, 0, 0, 0, 1), (1, 0)])

        log, entities = transform.copies(layout, m=transform.Matrix44.scale(2, 3, 1))
        assert len(log) == 0
        assert len(entities) == 1
        assert entities[0].dxftype() == "ELLIPSE"


class TestMLeaderNonUniformScaling:
    @pytest.fixture(scope="class")
    def msp(self):
        from ezdxf.render import mleader
        from ezdxf.math import Vec2

        doc = ezdxf.new()
        msp = doc.modelspace()
        ml_builder = msp.add_multileader_mtext("Standard")
        ml_builder.set_content("Line1\nLine2")
        ml_builder.add_leader_line(mleader.ConnectionSide.right, [Vec2(40, 15)])
        ml_builder.build(insert=Vec2(5, 0))
        return msp

    def test_transformation_will_not_be_applied_inplace(self, msp):
        log = transform.scale(msp, 2, 3, 1)
        assert log[0].error == transform.Error.NON_UNIFORM_SCALING_ERROR
        assert msp[0].dxftype() == "MULTILEADER"

    def test_mleader_will_not_copied(self, msp):
        log, clones = transform.copies(msp, m=transform.Matrix44.scale(2, 3, 1))
        assert log[0].error == transform.Error.NON_UNIFORM_SCALING_ERROR
        assert len(clones) == 0


if __name__ == "__main__":
    pytest.main([__file__])
